// $Id: PDFViewer.java 10479 2009-07-10 09:51:07Z chris $

package org.faceless.pdf2.viewer;

import org.faceless.pdf2.*;
import javax.swing.*;
import javax.swing.event.*;
import java.io.*;
import java.awt.event.*;
import java.awt.*;
import javax.swing.filechooser.FileFilter;

/**
 * The <code>PDFViewer</code> class is a simple Swing PDF viewer application.
 * It demonstrates the {@link PDFScrollPane} class, and can be run directly
 * from the JAR (via the {@link PDFTool} class) like so.
 * <pre>
 * java -jar bfopdf.jar <i>filename</i>
 * </pre>
 * The <i>filename</i> argument is optional, but if supplied will load the
 * specified PDF.
 *
 * <p><i>
 * This code is copyright the Big Faceless Organization. You're welcome to
 * use, modify and distribute it in any form in your own projects, provided
 * those projects continue to make use of the Big Faceless PDF library.
 * </i></p>
 *
 * @since 2.5
 */
public class PDFViewer extends JPanel
{
    private static final int INITWIDTH = 400;
    private static final int INITHEIGHT = 200;
    private static final String[] percentage = {"12.5%", "25%", "50%", "75%", "100%", "125%", "150%",
                                           "200%", "400%", "800%", "1600%", "2400%", "3200%", "6400%"};

    static final ImageIcon BFO16 = new ImageIcon(PDFViewer.class.getResource("resources/BFO16.png"));

    // Components
    private JDesktopPane desktop;
    private JComboBox cboZoom;
    private JButton filePrintButton;
    private JMenuItem mCascade;
    private JMenuItem mTiff;
    private JPanel toolbarPanel;

    // Other bits
    private final JFrame parent;
    private int frameCounter = 0;
    private PDFDocumentFrame activeFrame;
    private JFileChooser chooser;

    /**
     * Creates new PDFViewer
     */
    PDFViewer(JFrame frame) {
        parent = frame;
        try {
            UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
        } catch (Exception e) { }

        setPreferredSize(new Dimension(500, 500));

        JPanel toolbarPanel = new javax.swing.JPanel();
        toolbarPanel.setBorder(BorderFactory.createEtchedBorder());
        JToolBar fileToolbar = new javax.swing.JToolBar();
        fileToolbar.setLayout(new FlowLayout(FlowLayout.LEFT, 3, 0));
        fileToolbar.setBorder(BorderFactory.createEmptyBorder());

        JButton fileOpenButton = new javax.swing.JButton("Open");
        fileOpenButton.setToolTipText("Open File...");
        filePrintButton = new javax.swing.JButton("Print");
        filePrintButton.setToolTipText("Print...");
        filePrintButton.setEnabled(false);

        JMenuBar mbPdf = new JMenuBar();
        JMenu mFile = new JMenu();
        JMenu mWindow = new JMenu();
        mCascade = new JMenuItem();
        mCascade.setEnabled(false);
        JMenu mTools = new JMenu();
        mTiff = new JMenuItem();
        mTiff.setEnabled(false);

        setLayout(new BorderLayout());
        toolbarPanel.setLayout(new java.awt.FlowLayout(java.awt.FlowLayout.LEFT, 0, 0));

        fileOpenButton.setPreferredSize(new Dimension(62, 32));
        fileOpenButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                loadPDF(null);
            }
        });
        fileToolbar.add(fileOpenButton);

        filePrintButton.setPreferredSize(new Dimension(62, 32));
        filePrintButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                activeFrame.print();
            }
        });
        fileToolbar.add(filePrintButton);

        cboZoom = new javax.swing.JComboBox(percentage);
        if (System.getProperty("os.name").toLowerCase().startsWith("mac os x")) {
            cboZoom.setFont(cboZoom.getFont().deriveFont(Font.PLAIN, 10));
        }
        cboZoom.setEnabled(false);
        cboZoom.setSelectedIndex(4);
        cboZoom.setPreferredSize(new Dimension(68, 32));
        cboZoom.setEditable(true);
        cboZoom.setToolTipText("Zoom Tool");
        cboZoom.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                final JComboBox cb = (JComboBox) evt.getSource();
                if (cb.getSelectedItem()!=null) {
                    activeFrame.setPercent((String)cb.getSelectedItem());
                }
            }
        });

        fileToolbar.add(cboZoom);
        toolbarPanel.add(fileToolbar);
        add(toolbarPanel, BorderLayout.NORTH);

        // add desktop pane
        desktop = new JDesktopPane();
        desktop.setBackground(Color.gray);
        add(desktop,BorderLayout.CENTER);

        mbPdf.setDoubleBuffered(true);
        mFile.setText("File");
        JMenuItem mFileOpen = new JMenuItem();
        mFileOpen.setText("Open");
        mFileOpen.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                loadPDF(null);
            }
        });
        mFile.add(mFileOpen);
        JMenuItem mFileQuit = new JMenuItem();
        mFileQuit.setText("Quit");
        mFileQuit.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                System.exit(0);
            }
        });
        mFile.add(mFileQuit);
        mbPdf.add(mFile);

        mTools.setText("Tools");
        mTiff.setText("PDF -> TIFF");
        mTiff.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                PDFViewport vp = activeFrame.getViewPort();
                TiffDialog tiff = new TiffDialog(parent, vp, true);
            }
        });
        mTools.add(mTiff);
        mbPdf.add(mTools);

        mWindow.setText("Window");
        mCascade.setText("Cascade");
        mCascade.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                cascadeFrames();
            }
        });
        mWindow.add(mCascade);

        JMenuItem mAbout = new JMenuItem();
        mAbout.setText("About");
        mAbout.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                JOptionPane.showInternalMessageDialog(desktop, "Big Faceless PDF Library Viewer\nVersion "+PDF.VERSION+"\n\njava -jar bfopdf.jar --help for more options", "About", JOptionPane.INFORMATION_MESSAGE);
            }
        });
        mWindow.add(mAbout);
        mbPdf.add(mWindow);

        parent.setJMenuBar(mbPdf);
        setVisible(true);

        new Thread() {
            public void run() {
                chooser = new JFileChooser((File)null);
                chooser.setFileFilter(new FileFilter() {
                    public boolean accept(File f) {
                        return f.isDirectory() || f.getName().toLowerCase().endsWith(".pdf");
                    }
                    public String getDescription() {
                        return "PDF files (*.pdf)";
                    }
                });
            }
        }.start();
    }

    /**
     * Load a PDF into the viewer from a file.
     * @param file the PDF file to load, or <code>null</code> to select it with a chooser
     * @since 2.7.1
     */
    public void loadPDF(File file) {
        loadPDF(file, null);
    }

    /**
     * Load a PDF into the viewer from a file. If a password is specified it will be used
     * to open the file (if it's null or incorrect, the user will be prompted).
     * @param file the PDF file to load, or <code>null</code> to select it with a chooser
     * @param password the password to use to open the file - may be <code>null</code>.
     * @since 2.7.1
     */
    public void loadPDF(File file, final String password) {
        if (file == null) {
            int returnVal = chooser.showOpenDialog(parent);
            if (returnVal == JFileChooser.APPROVE_OPTION) {
                repaint();
                file = chooser.getSelectedFile();
            }
        }

        if (file!=null) {
            final File ffile = file;
            javax.swing.SwingUtilities.invokeLater(new Runnable() {
                public void run() {
                    try {
                        PDFReader reader = null;
                        try {
                            reader = new PDFReader(ffile, password);
                        } catch (PasswordException pwd) {
                            PasswordDialog login = new PasswordDialog(parent,ffile);
                            repaint();
                            if (login.isValidPassword()) {
                                reader = login.getReader();
                            }
                        }
                        if (reader!=null) {
                            addDocumentFrame(new PDF(reader), ffile.getName());
                        }
                    } catch (Exception e) {
                        activeFrameChanged();
                        ErrorInfo.displayThrowable(e, "Open Document Error", parent);
                    }
                }
            });
        }
    }

    /**
     * Load a pre-loaded PDF into the viewer.
     * @param pdf the PDF to load
     * @param name the name of the PDF, to display in the title bar.
     * @since 2.7.1
     */
    public void loadPDF(final PDF pdf, final String name) {
        if (pdf!=null) {
            javax.swing.SwingUtilities.invokeLater(new Runnable() {
                public void run() {
                    addDocumentFrame(pdf, name==null ? pdf.getInfo("Title") : name);
                }
            });
        }
    }

    private void addDocumentFrame(PDF pdf, String name) {
        PDFDocumentFrame frame = new PDFDocumentFrame();
        frame.setFrameIcon(BFO16);
        frame.setBounds(frameCounter*30, frameCounter*20, INITWIDTH, INITHEIGHT);
        frameCounter = (frameCounter+1) % 10;
        frame.addInternalFrameListener(frame. new FrameListener());
        try {
            // minimise other frames
            JInternalFrame[] fs = desktop.getAllFrames();
            for (int i=0;i<fs.length;i++) fs[i].setMaximum(false);
            // add new frame
            desktop.add(frame);
            frame.show();
            frame.setMaximum(true);
        } catch (java.beans.PropertyVetoException e) { }
        frame.setTitle(name);
        frame.setDocument(pdf);
    }

    private void activeFrameChanged() {
        activeFrame = (PDFDocumentFrame)desktop.getSelectedFrame();
        mCascade.setEnabled(activeFrame != null);
        mTiff.setEnabled(activeFrame != null);
        if(activeFrame != null) {
            cboZoom.setSelectedItem(activeFrame.getPercent());
        }
        cboZoom.setEnabled(activeFrame != null);
        filePrintButton.setEnabled(activeFrame != null);
    }

    private void cascadeFrames() {
        try {
            JInternalFrame[] frames = desktop.getAllFrames();
            JInternalFrame selectedFrame = desktop.getSelectedFrame();
            int x = 0;
            int y = 0;
            for (int k=frames.length-1; k>=0; k--) {
                frames[k].setMaximum(false);
                frames[k].setIcon(false);
                frames[k].setBounds(x, y, INITWIDTH, INITHEIGHT);
                x += 20;
                y += 20;
            }
            if (selectedFrame != null) {
                desktop.setSelectedFrame(selectedFrame);
            }
        } catch(java.beans.PropertyVetoException e) { }
    }

    private class PDFDocumentFrame extends JInternalFrame implements PDFViewport.ViewportObserver {
        private javax.swing.JFormattedTextField tfPages;
        private javax.swing.JSplitPane splitPane;
        private javax.swing.JTabbedPane tabbedPane;
        private javax.swing.JPanel navigatePanel;
        private javax.swing.JButton btnMoveFirst;
        private javax.swing.JButton btnMoveLast;
        private javax.swing.JButton btnMoveLeft;
        private javax.swing.JButton btnMoveRight;

        private PDFScrollPane pdfPanel;
        private PDFViewport pdfViewport;
        private OutlinePanel pdfOutlinePanel;
        private ThumbsPanel pdfThumbsPanel;
        private PDF document;
        private int currentPage;
        private JScrollBar verticalSB, horizontalSB;

        PDFDocumentFrame() {
            super("", true, true, true, true);

            currentPage = 0;
            setBackground(Color.gray);
            pdfPanel = new PDFScrollPane();
            pdfViewport = pdfPanel.getViewport();

            // Add some Rendering Hints - maybe they'll help?
            RenderingHints hints = new RenderingHints(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_QUALITY);
            hints.put(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BICUBIC);
            hints.put(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
            hints.put(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
            pdfViewport.setRenderingHints(hints);

            // add navigation panel
            navigatePanel = new javax.swing.JPanel();
            navigatePanel.setBorder(BorderFactory.createEtchedBorder());

            btnMoveFirst = new javax.swing.JButton("| <");
            btnMoveLeft = new javax.swing.JButton("<");
            btnMoveRight = new javax.swing.JButton(">");
            btnMoveLast = new javax.swing.JButton("> |");

            // add tabbed pane
            tabbedPane = new JTabbedPane(JTabbedPane.LEFT);

            // add document outline view
            pdfOutlinePanel = new OutlinePanel(pdfViewport);
            pdfOutlinePanel.setOpaque(true);
            tabbedPane.addTab("Bookmarks", null, pdfOutlinePanel, "Bookmarks");

            // add document thumbnail view
            pdfThumbsPanel = new ThumbsPanel(pdfViewport);
            tabbedPane.addTab("Pages", null, pdfThumbsPanel, "Pages");
            tabbedPane.addChangeListener(new ChangeListener() {
                // This method is called whenever the selected tab changes
                public void stateChanged(ChangeEvent evt) {
                    JTabbedPane pane = (JTabbedPane) evt.getSource();
                    // Get current tab
                    if (pane.getSelectedIndex() == 1) {
                        pdfThumbsPanel.setDocument(document);
                    }
                }
            });
            tabbedPane.setSelectedIndex(0);

            // add split pane
            splitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, tabbedPane, pdfPanel);
            splitPane.setOneTouchExpandable(true);
            splitPane.setDividerLocation(0);
            splitPane.setDividerSize(10);
            splitPane.setResizeWeight(0);
            splitPane.setContinuousLayout(true);
            getContentPane().add(splitPane, BorderLayout.CENTER);

            // add page buttons
            btnMoveFirst.addActionListener(new java.awt.event.ActionListener() {
                public void actionPerformed(java.awt.event.ActionEvent evt) {
                    currentPage = 0;
                    pdfViewport.setPageNumber(currentPage);
                }
            });
            navigatePanel.add(btnMoveFirst);

            btnMoveLeft.addActionListener(new java.awt.event.ActionListener() {
                public void actionPerformed(java.awt.event.ActionEvent evt) {
                    currentPage--;
                    pdfViewport.setPageNumber(currentPage);
                }
            });
            navigatePanel.add(btnMoveLeft);

            tfPages = new javax.swing.JFormattedTextField();
            tfPages.setPreferredSize(new Dimension(70, 21));
            tfPages.setHorizontalAlignment(JTextField.CENTER);
            tfPages.setEditable(true);
            tfPages.addKeyListener(new KeyAdapter() {
                public void keyPressed(KeyEvent e) {
                    if (e.getKeyCode() == KeyEvent.VK_ENTER) {
                        try {
                            String s = tfPages.getText();
                            if (s.indexOf("of")>0) {
                                s = s.substring(0, s.indexOf("of")-1);
                            }
                            s = s.trim();
                            int pgnum = Integer.parseInt(s) - 1;
                            if (pgnum<0 || pgnum >= document.getNumberOfPages() || pgnum==currentPage) {
                                throw new RuntimeException();
                            }
                            currentPage = pgnum;
                            pdfViewport.setPageNumber(currentPage);
                        } catch (Exception x) {
                            tfPages.setText((currentPage+1) + " of " + document.getNumberOfPages());
                        }
                    }
                }
            });
            tfPages.addFocusListener(new FocusAdapter() {
                public void focusGained(FocusEvent e) {
                    tfPages.selectAll();
                }
            });
            navigatePanel.add(tfPages);

            btnMoveRight.addActionListener(new java.awt.event.ActionListener() {
                public void actionPerformed(java.awt.event.ActionEvent evt) {
                    currentPage++;
                    pdfViewport.setPageNumber(currentPage);
                }
            });
            navigatePanel.add(btnMoveRight);

            btnMoveLast.addActionListener(new java.awt.event.ActionListener() {
                public void actionPerformed(java.awt.event.ActionEvent evt) {
                    currentPage = document.getNumberOfPages() - 1;
                    pdfViewport.setPageNumber(currentPage);
                }
            });
            navigatePanel.add(btnMoveLast);

            navigatePanel.setVisible(false);
            getContentPane().add(navigatePanel, BorderLayout.SOUTH);
            setTitle("BFO - [" + title + "]");
            splitPane.setVisible(true);
            navigatePanel.setVisible(true);

            // add observers
            pdfViewport.addObserver(pdfThumbsPanel);
            pdfViewport.addObserver(this);
        }

        /**
         * Set the PDF for this frame
         */
        void setDocument(PDF pdf) {
//          add document to view
            this.document = pdf;
            pdfViewport.setDocument(document);
            pdfOutlinePanel.setDocument(document);
        }

        /**
         * Set the viewing percentage for this frame
         */
        void setPercent(String p) {
            if (p.indexOf("%")<0) {
                p = p+"%";
            }
            if (document!=null) {
                if (p.indexOf("%")>=0) {
                    p = p.substring(0, p.lastIndexOf("%"));
                }
                p = p.trim();
                try {
                    float dpi = Float.parseFloat(p) * 0.72f;
                    pdfViewport.setDPI(dpi);
                } catch (Exception e) { }
            }
        }

        public String getPercent() {
            return ((int)Math.round(pdfViewport.getDPI() / 72 * 100))+"%";
        }

        public void print() {
            if (document != null) pdfViewport.print();
        }

        public PDFViewport getViewPort() {
            return pdfViewport;
        }

        /* (non-Javadoc)
         * @see org.faceless.pdf2.viewer.PDFViewportObserver#viewportChanged(double, double)
         */
        public void viewportChanged(double x, double y) {
            if (document != null) {
                currentPage = pdfViewport.getPageNumber();
                btnMoveRight.setEnabled(currentPage != document.getNumberOfPages()-1);
                btnMoveLast.setEnabled(currentPage != document.getNumberOfPages()-1);
                btnMoveLeft.setEnabled(currentPage!=0);
                btnMoveFirst.setEnabled(currentPage!=0);
                tfPages.setText((currentPage+1) + " of " + document.getNumberOfPages());
                cboZoom.setSelectedItem(getPercent());
            }
        }

        class FrameListener extends InternalFrameAdapter {
            public void internalFrameClosing(InternalFrameEvent e) {
                dispose();
                desktop.remove(e.getInternalFrame());
                frameCounter = (frameCounter-1) % 10;
                activeFrameChanged();
            }

            public void internalFrameActivated(InternalFrameEvent e) {
                activeFrameChanged();
            }
        }

        /* (non-Javadoc)
         * @see org.faceless.pdf2.viewer.PDFViewport.ViewportObserver#writeAsTIFFComplete()
         */
        public void writeAsTIFFComplete() {
            JOptionPane.showMessageDialog(this, "TIFF created successfully", "PDF to TIFF", JOptionPane.INFORMATION_MESSAGE);
        }
    }

    /**
     * Create a new PDFViewer object in a frame of it's own. Top level routine
     * to be called by main()
     * @since 2.7..1
     */
    public static PDFViewer newPDFViewer() {
        JFrame bfoFrame = new JFrame("BFO");
        bfoFrame.setIconImage(BFO16.getImage());
        PDFViewer pdfViewer = new PDFViewer(bfoFrame);
        bfoFrame.addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) {
                System.exit(0);
            }
        });

        bfoFrame.getContentPane().add("Center", pdfViewer);
        bfoFrame.pack();

        // position window in the centre of the screen
        Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
        Dimension windowSize = bfoFrame.getSize();
        bfoFrame.setLocation(Math.max(0, (screenSize.width - windowSize.width) / 2), Math.max(0, (screenSize.height - windowSize.height) / 2));

        bfoFrame.setVisible(true);
        return pdfViewer;
    }

    /**
     * The main() method can be invoked to run this class from the command line.
     * A single argument specifying the name of the file to open is optional
     */
    public static void main(final String args[]) {
        File file = null;
        if (args.length > 0) {
            try {
                if (args.length > 3 || args[0].equals("-help")) {
                    System.err.println("Usage: java org.faceless.pdf2.viewer.PDFViewer [<file>]");
                    return;
                } else {
                    // load PDF
                    if (args.length == 1 || args.length == 3) {
                        String filename = args[args.length-1];
                        file = new File(filename);

                        if (!file.isFile() || !file.canRead()) {
                             System.err.println("PDFViewer: Unable to read \""+filename+"\"");
                             file = null;
                        }
                    }

                    if (args[0].equals("-license")) {
                        PDF.setLicenseKey(args[1]);
                    }
                }
            } catch (Exception e) {
                System.err.println("Usage: java org.faceless.pdf2.viewer.PDFViewer [<file>]");
                return;
            }
        }

        final File ffile = file;
        //Schedule a job for the event-dispatching thread:
        //creating and showing this application's GUI.
        javax.swing.SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                PDFViewer viewer = newPDFViewer();
                if (ffile != null) viewer.loadPDF(ffile);
            }
        });
    }
}
