// $Id: PDFViewer.java 15952 2012-06-28 16:57:17Z mike $

package org.faceless.pdf2.viewer2;

import org.faceless.pdf2.*;
import org.faceless.util.ToolBarLayout;
import org.faceless.pdf2.viewer2.feature.*;
import org.faceless.pdf2.viewer2.util.KeyStoreTrustManager;
import org.faceless.pdf2.viewer2.util.AppletMenuItem;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.border.*;
import java.io.*;
import java.net.URL;
import java.awt.event.*;
import java.awt.*;
import java.util.*;
import java.security.*;
import java.util.Timer;
import java.util.List;
import java.util.prefs.Preferences;
import java.beans.PropertyChangeEvent;
import java.beans.VetoableChangeListener;
import java.beans.PropertyVetoException;
import java.beans.PropertyChangeListener;
import javax.swing.filechooser.FileFilter;
import javax.swing.plaf.metal.*;
import javax.swing.plaf.*;

/**
 * The <code>PDFViewer</code> class is a simple Swing PDF viewer application.
 * It demonstrates the {@link DocumentPanel} class, and can be run directly
 * from the JAR (via the {@link PDFTool} class) like so.
 * <pre>
 * java -jar bfopdf.jar <i>filename</i>
 * </pre>
 * The <i>filename</i> argument is optional, but if supplied will load the
 * specified PDF.
 * See the <a href="doc-files/tutorial.html">viewer tutorial</a> for more detail on how to use this class and the "viewer" package.
 *
 * <div class="initparams">
 * The following <a href="doc-files/initparams.html">initialization parameters</a> may be specified
 * <table>
 * <tr><th>currentUser</th><td>The name of the current user - optional, but if set this will be set as the author field on any annotations created by the user</td></tr>
 * <tr><th>unpromptedDirtyClose</th><td>Whether to prompt the user for confirmation when closing a window containng a PDF that has been {@link DocumentPanel#setDirty modified}</td></tr>
 * <tr><th>dpi</th><td>The resolution of the screen. This can be set to override {@link Toolkit#getScreenResolution} if required</td></tr>
 * </table>
 * <p><i>This code is copyright the Big Faceless Organization. You're welcome to use, modify and distribute it in any form in your own projects, provided those projects continue to make use of the Big Faceless PDF library.</i></p>
 *
 * @since 2.5, with major rewrites in 2.8
 */
public class PDFViewer extends JPanel {

    static final ImageIcon BFO16 = new ImageIcon(PDFViewer.class.getResource("resources/BFO16.png"));
    private static final int MINIMIZEDWIDTH = 600;
    private static final int MINIMIZEDHEIGHT = 400;
    private JFileChooser filechooser;

    // Components
    private JDesktopPane desktop;     // Used when MultiWindow is set
    private JMenuBar menubar;
    private Map components;           // Map containing all components
    private Map toolbars;             // Map containing all the Toolbars that can be activated
    private Map menus;                // Map containing all the Menus
    private Collection pdfcomponents; // Components that require a PDF to be active
    private Collection listeners;     // Collection of DocumentPanelListeners

    // Other bits
    private final Collection features;
    private int internalFrameCounter = 0;
    private DocumentPanel activedocument;
    private KeyStoreManager keystoremanager;
    private PropertyManager propertymanager;
    private JSManager jsmanager;
    private LinkedHashMap aboutkeys;
    private Preferences preferences;
    private boolean createdEventPending; // For lazy create event on non MultiWindow viewers
    private boolean initialized, unprompteddirtyclose;
    private String user;

    /**
     * Creates new PDFViewer with the specified features
     * @param features a Collection of {@link ViewerFeature} objects that are enabled
     * @see ViewerFeature
     * @since 2.7.9
     */
    public PDFViewer(Collection features) {
        Util.initialize();
        this.features = Collections.unmodifiableCollection(features);
        setPropertyManager(new PropertyManager() {
            public String getProperty(final String key) {
                try {
                    return (String)AccessController.doPrivileged(new PrivilegedAction() {
                        public Object run() {
                            String x = System.getProperty("org.faceless.pdf2.viewer2."+key);
                            if (x==null) x = System.getProperty(key);
                            return x;
                        }
                    });
                } catch (SecurityException e) {
                    return null;
                }
            }
            public URL getURLProperty(String key) throws java.net.MalformedURLException {
                String val = getProperty(key);
                return val==null ? null : new URL(val);
            }
        });
        try {
            preferences = Preferences.userNodeForPackage(PDFViewer.class);
        } catch (Exception e) { }
        components = new LinkedHashMap();
        toolbars = new LinkedHashMap();
        menus = new LinkedHashMap();
        pdfcomponents = new HashSet();
        listeners = new LinkedHashSet();
        try {
            AccessController.doPrivileged(new PrivilegedAction() {
                public Object run() {
                    filechooser = new JFileChooser((File)null);
                    return null;
                }
            });
        } catch (SecurityException e) { }

        setLayout(new BorderLayout());
        setOpaque(true);

        addHierarchyListener(new HierarchyListener() {
            public void hierarchyChanged(HierarchyEvent event) {
                removeHierarchyListener(this);
                initialize();
            }
        });
        try {
            user = System.getProperty("user.name");
        } catch (Throwable e) { }
    }

    private synchronized void initialize() {
        if (initialized) return;
        initialized = true;

        getMenu("File");           // Initialize the order
        getMenu("Edit");
        getMenu("View");
        getMenu("Document");
        if (hasFeature(MultiWindow.getInstance())) {                    // This has to go in first
            addDocumentPanelListener(new DocumentPanelListener() {
                public void documentUpdated(DocumentPanelEvent e) {
                    if (e.getType()=="activated" || e.getType()=="deactivated") {
                        boolean enable = e.getType()=="activated";
                        for (Iterator i = pdfcomponents.iterator();i.hasNext();) {
                            JComponent comp = (JComponent)i.next();
                            comp.setEnabled(enable);
                        }
                    }
                }
            });
        }
        for (Iterator i = features.iterator();i.hasNext();) {
            Object o = i.next();
            if (o instanceof ViewerFeature) {
                ((ViewerFeature)o).initialize(this);
            }
            if (o instanceof Importer && filechooser!=null) {
                filechooser.addChoosableFileFilter(((Importer)o).getFileFilter());
            }
        }

        if (hasFeature(Toolbars.getInstance())) {
            // Ensure items added to this toolbarPanel are added in the order defined in
            // the "toolbars" map. Why? So if we float a toolbar then unfloat it, it drops
            // back into the correct position. Why not use setComponentZOrder() on a
            // ancestor listener? Because those events don't seem to fire in a predictable
            // order. This is more predictable, and anyway it works under Java 1.4.
            JPanel toolbarPanel = new JPanel(new ToolBarLayout()) {
                protected void addImpl(Component comp, Object constraints, int index) {
                    int pos = 0;
                    for (Iterator i = toolbars.values().iterator();i.hasNext();pos++) {
                        if (i.next()==comp) {
                            index = pos >= getComponentCount() ? -1 : pos;
                            break;
                        }
                    }
                    super.addImpl(comp, constraints, index);
                }
            };
            toolbarPanel.setFont(toolbarPanel.getFont().deriveFont(10f));
            for (Iterator i = toolbars.entrySet().iterator();i.hasNext();) {
                Map.Entry e = (Map.Entry)i.next();
                String key = (String)e.getKey();
                JComponent toolbar = (JComponent)e.getValue();
                toolbar.setFont(null);
                if (toolbar instanceof JToolBar) {
                    // If we only have one "ViewMode", no point in displaying toolbar
                    int minbuttons = "ViewMode".equals(key) ? 1 : 0;
                    if (toolbar.getComponentCount()>minbuttons) {
                        toolbarPanel.add(toolbar);
                    }
                } else {
                    add(toolbar);
                }
            }
            if (toolbarPanel.getComponentCount() > 0) {
                int bw = Util.isLAFWindows() ? 1 : 0;
                toolbarPanel.setBorder(BorderFactory.createEmptyBorder(bw, bw, bw, bw));
                toolbarPanel.setOpaque(false);
                add(toolbarPanel, BorderLayout.NORTH);
            }
        }

        if (hasFeature(Menus.getInstance())) {
            menubar = new JMenuBar();
            menubar.setDoubleBuffered(true);

            if (hasFeature(Toolbars.getInstance()) && hasFeature(ToolbarDisabling.getInstance())) {
                JMenuItem mToolbars = new JMenu(UIManager.getString("PDFViewer.Toolbars"));
                for (Iterator i = toolbars.entrySet().iterator();i.hasNext();) {
                    Map.Entry e = (Map.Entry)i.next();
                    final String key = (String)e.getKey();
                    final JComponent toolbar = (JComponent)e.getValue();
                    if (!key.endsWith(" ")) {
                        final JCheckBoxMenuItem toolbaritem = new JCheckBoxMenuItem();
                        mToolbars.add(toolbaritem);
                        String name = UIManager.getString("PDFViewer."+key);
                        if (name==null) {
                            if (getPropertyManager().getProperty("debug.L10N")!=null) {
                                System.err.println("No localization for menu: PDFViewer."+key);
                            }
                            name = key;
                        }
                        toolbaritem.setText(name);
                        toolbaritem.setSelected(toolbar.isVisible());
                        toolbaritem.addActionListener(new ActionListener() {
                            public void actionPerformed(ActionEvent event) {
                                toolbar.setVisible(toolbaritem.getState());
                            }
                        });
                    }
                }
                getMenu("View").add(mToolbars, 0);
            }

            if (hasFeature(MultiWindow.getInstance())) {
                JMenuItem item = setMenu("Window\tCascade", (char)0, false, new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        cascadeFrames();
                    }
                });
                components.put("MenuCascade", item);
            }

            for (Iterator i = menus.entrySet().iterator();i.hasNext();) {
                Map.Entry e = (Map.Entry)i.next();
                String key = (String)e.getKey();
                JMenu menu = (JMenu)e.getValue();
                if (menu.getItemCount()>0) {
                    menubar.add(menu);
                }
            }

            if (menubar.getComponentCount() > 0) {
                // Add the menubar to the parenet frame when the frame is set
                addPropertyChangeListener(new PropertyChangeListener() {
                    public void propertyChange(PropertyChangeEvent event) {
                        if ("ancestor".equals(event.getPropertyName())) {
                            boolean browserapplet = Util.isBrowserApplet(PDFViewer.this);
                            JApplet applet = (JApplet)SwingUtilities.getAncestorOfClass(JApplet.class, PDFViewer.this);
                            JFrame frame = (JFrame)SwingUtilities.getAncestorOfClass(JFrame.class, PDFViewer.this);
                            if (applet != null) {
                                applet.setJMenuBar(menubar);
                                List l = new ArrayList();           // remove all mnemonics
                                for (int i=0;i<menubar.getMenuCount();i++) {
                                    l.add(menubar.getMenu(i));
                                }
                                for (int i=0;i<l.size();i++) {
                                    JMenuItem item = (JMenuItem)l.get(i);
                                    if (item instanceof AppletMenuItem) {
                                        ((AppletMenuItem)item).setBrowserApplet(browserapplet);
                                    } else if (item instanceof JMenu) {
                                        for (int j=0;j<((JMenu)item).getItemCount();j++) {
                                            l.add(((JMenu)item).getItem(j));
                                        }
                                    }
                                }
                            } else if (frame != null) {
                                frame.setJMenuBar(menubar);
                            }
                        }
                    }
                });
            }
        }

        if (hasFeature(MultiWindow.getInstance())) {
            desktop = new JDesktopPane();
            desktop.setPreferredSize(new Dimension(600, 400));
            desktop.setOpaque(true);
            desktop.setBackground(Color.gray);
            add(desktop, BorderLayout.CENTER);

        } else {
            activedocument = new DocumentPanel();
            createdEventPending = true;
            activedocument.setViewer(this);
            activedocument.setFocusable(true);
            for (Iterator i = features.iterator();i.hasNext();) {
                ViewerFeature feature = (ViewerFeature)i.next();
                if (feature instanceof SidePanelFactory) {
                    activedocument.addSidePanelFactory((SidePanelFactory)feature);
                } else if (feature instanceof AnnotationComponentFactory) {
                    activedocument.addAnnotationComponentFactory((AnnotationComponentFactory)feature);
                } else if (feature instanceof ActionHandler) {
                    activedocument.addActionHandler((ActionHandler)feature);
                }
            }
            activedocument.setPreferredSize(new Dimension(600, 400));
            add(activedocument, BorderLayout.CENTER);
        }
        updatePropertyManager();
    }

    /**
     * Return the {@link Preferences} object that should be used to store user preferences
     * about the viewer, or <code>null</code> if preferences cannot be saved.
     * @since 2.11
     */
    public Preferences getPreferences() {
        return preferences;
    }

    /**
     * Return the JSManager object for this PDFViewer.
     * @since 2.9
     */
    public synchronized JSManager getJSManager() {
        initialize();
        if (jsmanager==null) {
            // Create a JSManager and initialize it as soon as possible, whether application is displayed or not
            jsmanager = new JSManager(this);
            jsmanager.runEventAppInit();
        }
        return jsmanager;
    }

    /**
     * Set the {@link PropertyManager} in use by this PDFViewer
     * @since 2.8.5
     */
    public final void setPropertyManager(PropertyManager manager) {
        if (manager==null) {
            throw new NullPointerException();
        }
        this.propertymanager = manager;
        updatePropertyManager();
    }

    private void updatePropertyManager() {
        String val = propertymanager.getProperty("currentUser");
        if (user != null) {
            setCurrentUser(user);
        }
        val = propertymanager.getProperty("unpromptedDirtyClose");
        unprompteddirtyclose = val != null && !"false".equals(val);
    }

    /**
     * Get the {@link PropertyManager} in use by this PDFViewer
     * @since 2.8.5
     */
    public final PropertyManager getPropertyManager() {
        return this.propertymanager;
    }

    /**
     * Get the {@link KeyStoreManager} in use by this PDFViewer
     * @since 2.8.3
     */
    public synchronized KeyStoreManager getKeyStoreManager() {
        initialize();
        if (keystoremanager==null) {
            keystoremanager = new KeyStoreManager(this);
            try {
                // Check legacy properties first
                String[] s = new String[] { "type", "provider", "file", "password" };
                for (int i=0;i<s.length;i++) {
                    String val = getPropertyManager().getProperty("KeyStoreManager."+s[i]);
                    if (val != null) {
                        keystoremanager.setParameter(s[i], val);
                    }
                }
                String val = getPropertyManager().getProperty("KeyStoreManager.params");
                if (val != null) {
                    keystoremanager.setParameters(val);
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
        return keystoremanager;
    }

    /**
     * Set the {@link KeyStoreManager} used by this PDFViewer
     * @since 2.8.3
     */
    public synchronized void setKeyStoreManager(KeyStoreManager manager) {
        this.keystoremanager = manager;
    }

    /**
     * Return true if the Viewer has a certain feature enabled.
     * @param feature the feature to look for
     */
    public boolean hasFeature(ViewerFeature feature) {
        initialize();
        return features.contains(feature);
    }

    /**
     * Return true if the Viewer has a certain feature enabled.
     * @param feature the name of the feature to look for
     */
    public boolean hasFeature(String feature) {
        initialize();
        for (Iterator i = features.iterator();i.hasNext();) {
            if (((ViewerFeature)i.next()).getName().equals(feature)) return true;
        }
        return false;
    }

    /**
     * Return the list of all features set in the viewer
     */
    public ViewerFeature[] getFeatures() {
        initialize();
        return (ViewerFeature[])new ArrayList(features).toArray(new ViewerFeature[0]);
    }

    /**
     * Return the specified <code>ViewerFeature</code>, or <code>null</code>
     * if it doesn't exist for this viewer
     * @since 2.8.5
     */
    public ViewerFeature getFeature(String feature) {
        initialize();
        for (Iterator i = features.iterator();i.hasNext();) {
            ViewerFeature f = (ViewerFeature)i.next();
            if (f.getName().equals(feature)) {
                return f;
            }
        }
        return null;
    }

    /**
     * Return the specified <code>ViewerFeature</code>, or <code>null</code>
     * if it doesn't exist for this viewer
     * @param clazz the Class of the ViewerFeature.
     * @since 2.10.2
     */
    public ViewerFeature getFeature(Class clazz) {
        initialize();
        for (Iterator i = features.iterator();i.hasNext();) {
            ViewerFeature f = (ViewerFeature)i.next();
            if (clazz.isInstance(f)) return f;
        }
        return null;
    }

    //-------------------------------------------------------------------------------
    // Methods to be called by Features when they're adding themselves to this Viewer
    //-------------------------------------------------------------------------------

    /**
     * Return the specifeid ToolBar. If the toolbar doesn't exist, it's created.
     * @param name the name of the ToolBar
     * @param enabled whether the toolbar is enabled or not
     * @param always whether the toolbar is always in it's enabled state
     */
    final JComponent getToolBar(String name, boolean enabled, boolean enabledalways, boolean floatable, boolean floating) {
        name = name.trim();
        if (enabledalways) {
            name = name + " ";
        }
        if (floating) {
            final String nname = name;
            final JInternalFrame toolbar = new JInternalFrame(name, false, true);
            if (hasFeature(ToolbarDisabling.getInstance())) {
                toolbar.addInternalFrameListener(new InternalFrameListener() {
                    public void internalFrameClosing(InternalFrameEvent e) {
                        JMenu menu = (JMenu)getMenu("View").getItem(0);
                        for (int i=0;i<menu.getItemCount();i++) {
                            if (nname.equals(menu.getItem(i).getText())) {
                                ((JCheckBoxMenuItem)menu.getItem(i)).setState(false);
                            }
                        }
                        toolbar.dispose();
                    }
                    public void internalFrameActivated(InternalFrameEvent e) { }
                    public void internalFrameDeactivated(InternalFrameEvent e) { }
                    public void internalFrameDeiconified(InternalFrameEvent e) { }
                    public void internalFrameIconified(InternalFrameEvent e) { }
                    public void internalFrameOpened(InternalFrameEvent e) { }
                    public void internalFrameClosed(InternalFrameEvent e) { }
                });
            }
            toolbar.setLocation(100+(toolbars.size()*20), 60+(toolbars.size()*20));
            toolbar.setVisible(enabled);
            toolbars.put(name, toolbar);
            return toolbar;
        } else {
            JToolBar toolbar = (JToolBar)toolbars.get(name);
            if (toolbar==null) {
                toolbar = new JToolBar();
                toolbar.setLayout(new GridBagLayout() {
                    { defaultConstraints.fill = GridBagConstraints.BOTH; defaultConstraints.weighty = 1; }
                });
                toolbar.setBorder(BorderFactory.createEtchedBorder());
                toolbar.setFloatable(floatable && !Util.isLAFNimbus() && !Util.isLAFGTK() && hasFeature(ToolbarFloating.getInstance()));
                toolbar.setVisible(enabled);
                toolbar.setOpaque(false);
                toolbars.put(name, toolbar);
            }
            return toolbar;
        }
    }

    /**
     * Add or replace a menu item in the viewer.
     * @param name the name of the menu, if the form "File\tOpen" or "File\tQuit(999)". The tab character is used to separate
     * the items in the menu hierarchy, and the optional bracked value is used to control ordering in the menu.
     * @param mnemonic the mnemonic to assign to this menu - lowercase or uppercase letter, or <code>(char)0</code> for no mnenonic
     * @param documentrequired whether this menu item should only be enabled if a document is loaded
     * @param listener the ActionListener to run when this menu item is activated
     * @since 2.10.2
     */
    public JMenuItem setMenu(String name, char mnemonic, boolean documentrequired, ActionListener listener) {
        String[] sections = name.split("\t");
        int[] sorts = new int[sections.length];
        for (int i=1;i<sections.length;i++) {
            String section = sections[i];
            sorts[i] = 10;
            int pos = section.lastIndexOf("(");
            if (pos >= 0 && section.endsWith(")")) {
                try {
                    sorts[i] = Integer.parseInt(section.substring(pos+1, section.length()-1));
                    sections[i] = section = section.substring(0, pos);
                } catch (RuntimeException e) { }
            }
            String sectiontext = UIManager.getString("PDFViewer."+section);
            if (sectiontext==null) {
                if (getPropertyManager().getProperty("debug.L10N")!=null) {
                    System.err.println("No localization for menu: PDFViewer."+section);
                }
            } else {
                sections[i] = sectiontext;
            }
        }

        JMenu menu = getMenu(sections[0]);
        JMenuItem item = null;
        for (int i=1;i<sections.length;i++) {
            item = null;
            final String section = sections[i];
            final int sort = sorts[i];
            int pos = -1;
            for (int j=0;item==null && j<menu.getItemCount();j++) {
                item = menu.getItem(j);
                if (item != null && item.getText() != null && !item.getText().equals(section)) {
                    Integer oldsort = (Integer)item.getClientProperty("bfo.sort");
                    if (oldsort!=null && oldsort.intValue() > sort) {
                        pos = j;
                    }
                    item = null;
                }
            }
            if (item==null) {
                if (i==sections.length-1) {
                    item = new AppletMenuItem(section);
                    if (mnemonic!=0) {
                        // Although mnemonics don't work in applets, we're not
                        // in a hierarchy yet so can't test. Add now, remove
                        // it later if needed.
                        int mask = getToolkit().getMenuShortcutKeyMask();
                        if (Character.isUpperCase(mnemonic)) {
                            mask |= java.awt.Event.SHIFT_MASK;
                        }
                        item.setAccelerator(KeyStroke.getKeyStroke(Character.toUpperCase(mnemonic), mask, false));
                    }

                    if (documentrequired) {
                        item.setEnabled(false);
                        addDocumentDependentComponent(item);
                    }
                } else {
                    item = new JMenu(section);
                }
                item.putClientProperty("bfo.sort", new Integer(sort));
                menu.add(item, pos);
                item.addActionListener(listener);
            }
            if (i<sections.length-1) {
                menu = (JMenu)item;
            }
        }
        return item;
    }

    /**
     * Return the specified Menu. If the menu doesn't exist, it's created
     * @param name the name of the Menu, before it's been localized
     * @since 2.10.4
     */
    public JMenu getMenu(String name) {
        String text = UIManager.getString("PDFViewer."+name);
        if (text==null) {
            if (getPropertyManager().getProperty("debug.L10N")!=null) {
                System.err.println("No localization for menu: PDFViewer."+name);
            }
            text = name;
        }
        JMenu menu = (JMenu)menus.get(name);
        if (menu==null) {
            menu = new JMenu(text);
            menus.put(name, menu);
        }
        return menu;
    }

    /**
     * Add a named component to the viewers list
     * @since 2.10.2
     */
    public void putNamedComponent(String name, JComponent value) {
        initialize();
        components.put(name, value);
    }

    /**
     * Add a component to the list of those that require a valid PDF to be open
     */
    void addDocumentDependentComponent(JComponent component) {
        pdfcomponents.add(component);
    }

    /**
     * Return a Component created by a {@link ViewerFeature}
     */
    public JComponent getNamedComponent(String name) {
        return (JComponent)components.get(name);
    }

    /**
     * Get the Frame this Viewer is inside. Used for dialogs
     */
    Frame getFrame() {
        return JOptionPane.getFrameForComponent(this);
    }

    //-------------------------------------------------------------------------------
    // Methods for loading a PDF or changing the active PDF
    //-------------------------------------------------------------------------------

    /**
     * Load a PDF into the viewer from a file. This method will search the
     * {@link Importer} objects in order until it finds one that matches the file,
     * then will use that to load the PDF into the viewer. Prompting for passwords
     * etc. is left to the appropriate {@link Importer} - see the {@link PDFImporter}
     * class for details on the default behaviour.
     * @param ffile the PDF file to load, or <code>null</code> to select it with a chooser
     * @since 2.7.1
     */
    public void loadPDF(final File ffile) {
        loadPDF(ffile, 0);
    }

    /**
     * Load a PDF into the viewer from a file. This method will search the
     * {@link Importer} objects in order until it finds one that matches the file,
     * then will use that to load the PDF into the viewer. Prompting for passwords
     * etc. is left to the appropriate {@link Importer} - see the {@link PDFImporter}
     * class for details on the default behaviour.
     * @param ffile the PDF file to load, or <code>null</code> to select it with a chooser
     * @param pagenumber the initial page number to request. May be overridden
     * @since 2.11.10
     */
    public void loadPDF(final File ffile, final int pagenumber) {
        initialize();
        AccessController.doPrivileged(new PrivilegedAction() {
            public Object run() {
                File file = ffile;
                if (file == null) {
                    PDFImporter pdfimporter = (PDFImporter)getFeature(PDFImporter.class);
                    if (pdfimporter != null) {
                        filechooser.setFileFilter(pdfimporter.getFileFilter());
                    }
                    String dir = (preferences == null) ? null : preferences.get("lastDirectory", null);
                    if (dir == null && preferences != null) {
                        dir = preferences.get("File/dir", null);
                        if (dir != null) {
                            preferences.remove("File/dir");
                        }
                    }
                    if (dir != null) {
                        filechooser.setCurrentDirectory(new File(dir));
                    }
                    if (filechooser.showOpenDialog(PDFViewer.this) == JFileChooser.APPROVE_OPTION) {
                        repaint();
                        file = filechooser.getSelectedFile();
                        if (file != null) {
                            dir = file.getParent();
                            if (dir != null && preferences != null) {
                                preferences.put("lastDirectory", dir);
                            }
                        }
                    }
                }
                if (file != null) {
                    try {
                        boolean match = false;
                        IOException exception = null;
                        if (!file.exists()) {
                            throw new IOException("\""+file+"\": No such File");
                        }
                        for (Iterator i = features.iterator(); !match && i.hasNext();) {
                            ViewerFeature feature = (ViewerFeature)i.next();
                            if (feature instanceof Importer) {
                                Importer importer = (Importer)feature;
                                try {
                                    if (match = importer.matches(file)) {
                                        Importer.ImporterTask task = importer.getImporter(PDFViewer.this, file);
                                        task.setPageNumber(pagenumber);
                                        task.start(PDFViewer.this, UIManager.getString("PDFViewer.Loading"));
                                    }
                                } catch (IOException e) {
                                    exception = e;
                                    match = false;
                                }
                            }
                        }
                        if (!match) {
                            if (exception != null) {
                                throw exception;
                            } else {
                                throw new IOException("Cannot load this file as a PDF");
                            }
                        }
                    } catch (IOException e) {
                        Util.displayThrowable(e, getFrame());
                    }
                }
                return null;
            }
        });
    }

    /**
     * Load a PDF into the viewer from a file using the specified EncryptionHandler.
     * This method is now deprecated - the correct way to load a PDF with custom
     * EncryptionHandlers is to modify the {@link PDFImporter} passed in to the
     * PDF Constructor.
     * @param file the PDF File
     * @param handler the EncryptionHandler to use
     * @since 2.8
     * @deprecated this method has been superceded by the {@link PDFImporter} class
     */
    public void loadPDF(File file, EncryptionHandler handler) {
        initialize();
        loadPDF(file, new EncryptionHandler[] { handler });
    }

    /**
     * Load a PDF into the viewer from a file using the specified list of EncryptionHandler.
     * This method is now deprecated - the correct way to load a PDF with custom
     * EncryptionHandlers is to modify the {@link PDFImporter} passed in to the
     * PDF Constructor.
     * @param ffile the PDF File
     * @param handlers the list of EncryptionHandlers
     * @since 2.8
     * @deprecated this method has been superceded by the {@link PDFImporter} class
     */
    public void loadPDF(final File ffile, final EncryptionHandler[] handlers) {
        initialize();
        AccessController.doPrivileged(new PrivilegedAction() {
            public Object run() {
                File file = ffile;
                if (file==null && filechooser.showOpenDialog(PDFViewer.this)==JFileChooser.APPROVE_OPTION) {
                    repaint();
                    file = filechooser.getSelectedFile();
                }
                if (file!=null) {
                    PDFImporter importer = (PDFImporter)getFeature(PDFImporter.class);
                    if (importer==null) {
                        importer = new PDFImporter();
                    }
                    if (handlers!=null) {
                        Set storeset = importer.getEncryptionHandlers();
                        Set newset = new HashSet(Arrays.asList(handlers));
                        if (!newset.equals(storeset)) {
                            importer = new PDFImporter();
                            importer.getEncryptionHandlers().clear();
                            importer.getEncryptionHandlers().addAll(newset);
                        }
                    }
                    importer.getImporter(PDFViewer.this, file).start(PDFViewer.this, UIManager.getString("PDFViewer.Loading"));
                }
                return null;
            }
        });
    }

    /**
     * Load a PDF into the viewer from an InputStream. Note that this method will
     * return immediately, which means the calling thread <i>must not close</i>
     * the InputStream. Doing so will result in an Exception. The stream will be
     * closed by the reading thread when it is completed.
     * </p><p>
     * This method is now deprecated - the correct way to load a PDF from an
     * InputStream is to call the {@link Importer#getImporter(PDFViewer, InputStream, String, File)}
     * method. For example:
     * <pre class="example">
     * PDFImporter importer = (PDFImporter)viewer.getFeature(PDFImporter.class);
     * importer.getImporter(viewer, inputstream, title, file).start(viewer, "Loading");
     * </pre>
     * @param in the InputStream to load the PDF from
     * @param handlers the EncryptionHandlers to use to try to decrypt the PDF
     * @param title The name of the Window. May be null.
     * @param file If using a save dialog, the initial value to set the dialog to. May be null.
     * @deprecated this method has been superceded by the {@link PDFImporter} class
     * @since 2.8.3
     */
    public void loadPDF(InputStream in, EncryptionHandler[] handlers, String title, File file) {
        initialize();
        PDFImporter importer = (PDFImporter)getFeature(PDFImporter.class);
        if (importer==null) {
            importer = new PDFImporter();
        }
        if (handlers!=null) {
            Set storeset = importer.getEncryptionHandlers();
            Set newset = new HashSet(Arrays.asList(handlers));
            if (!newset.equals(storeset)) {
                importer = new PDFImporter();
                importer.getEncryptionHandlers().clear();
                importer.getEncryptionHandlers().addAll(newset);
            }
        }
        importer.getImporter(this, in, title, file).start(this, UIManager.getString("PDFViewer.Loading"));
    }

    /**
     * Load a pre-loaded PDF into the viewer. Simply calls
     * <code>loadPDF(new PDFParser(pdf), name)</code>.
     * @param pdf the PDF to load
     * @param name the name of the PDF, to display in the title bar.
     * @since 2.7.1
     */
    public void loadPDF(PDF pdf, String name) {
        if (pdf!=null) {
            loadPDF(new PDFParser(pdf), name);
        }
    }

    /**
     * Load a pre-loaded PDF into the viewer.
     * @param parser the PDFParser referencing the PDF to load
     * @param name the name of the PDF, to display in the title bar.
     * @since 2.11.3
     */
    public void loadPDF(PDFParser parser, String name) {
        loadPDF(parser, name, 0);
    }

    /**
     * Load a pre-loaded PDF into the viewer.
     * @param parser the PDFParser referencing the PDF to load
     * @param name the name of the PDF, to display in the title bar.
     * @param pagenumber the initial page to open the PDF at
     * @since 2.11.3
     */
    public void loadPDF(PDFParser parser, String name, int pagenumber) {
        initialize();
        File file = name == null ? null : new File(name);
        addDocumentPanel(parser, name==null ? parser.getPDF().getInfo("Title") : name, file, pagenumber);
    }

    /**
     * Add a new Document Frame containing the specified PDF. If we're using the
     * MultiWindow feature then this is done in a JInternalFrame inserted into
     * the desktop. If not, we just update the activedocument with the new
     * PDF.
     */
    void addDocumentPanel(final PDFParser parser, final String name, final File file, final int pagenumber) {
        initialize();
        Runnable r = new Runnable() {
            public void run() {
                if (hasFeature(MultiWindow.getInstance())) {
                    final DocumentPanel documentpanel = new DocumentPanel();
                    documentpanel.putClientProperty("file", file);
                    for (Iterator i = listeners.iterator();i.hasNext();) {
                        documentpanel.addDocumentPanelListener((DocumentPanelListener)i.next());
                    }
                    documentpanel.setViewer(PDFViewer.this);
                    documentpanel.raiseDocumentPanelEvent(DocumentPanelEvent.createCreated(documentpanel));
                    for (Iterator i = features.iterator();i.hasNext();) {
                        ViewerFeature feature = (ViewerFeature)i.next();
                        if (feature instanceof SidePanelFactory) {
                            documentpanel.addSidePanelFactory((SidePanelFactory)feature);
                        } else if (feature instanceof AnnotationComponentFactory) {
                            documentpanel.addAnnotationComponentFactory((AnnotationComponentFactory)feature);
                        } else if (feature instanceof ActionHandler) {
                            documentpanel.addActionHandler((ActionHandler)feature);
                        }
                    }

                    final JInternalFrame frame = new JInternalFrame(name, true, true, true, true);
                    frame.addVetoableChangeListener(new VetoableChangeListener() {
                        public void vetoableChange(PropertyChangeEvent e) throws PropertyVetoException {
                            if (e.getPropertyName().equals(JInternalFrame.IS_CLOSED_PROPERTY) && ((Boolean)e.getNewValue()).booleanValue()) {
                                if (!getUnpromptedDirtyClose() && documentpanel.isDirty()) {
                                    frame.toFront();
                                    if (JOptionPane.showConfirmDialog(PDFViewer.this, UIManager.getString("PDFViewer.ConfirmCloseText"), UIManager.getString("PDFViewer.Confirm"), JOptionPane.YES_NO_OPTION) == JOptionPane.NO_OPTION) {
                                        throw new PropertyVetoException("Close cancelled", e);
                                    }
                                }
                            }
                        }
                    });

                    frame.setContentPane(documentpanel);
                    frame.setFrameIcon(BFO16);
                    frame.setFocusable(false);
                    frame.addInternalFrameListener(new InternalFrameListener() {
                        private void activeWindowChanged() {
                            JInternalFrame activeWindow = desktop.getSelectedFrame();
                            if (activeWindow != null && activeWindow.isIcon()) {
                                activeWindow = null;
                            }
                            if (activedocument != null) {
                                activedocument.raiseDocumentPanelEvent(DocumentPanelEvent.createDeactivated(activedocument));
                            }
                            activedocument = activeWindow!=null ? (DocumentPanel)activeWindow.getContentPane() : null;
                            if (activedocument != null && activedocument.getPDF() != null) {
                                activedocument.raiseDocumentPanelEvent(DocumentPanelEvent.createActivated(activedocument));
                            }
                            updateWindowMenu(false);
                        }

                        public void internalFrameClosed(InternalFrameEvent e) {
                            frame.removeInternalFrameListener(this);
                            documentpanel.setPDF(null);
                            frame.removeAll();
                            frame.dispose();
                            desktop.remove(frame);
                            activeWindowChanged();
                            if (desktop.getComponentCount() == 0) {
                                internalFrameCounter = 0;
                                // Bug 4759312
                                KeyboardFocusManager.getCurrentKeyboardFocusManager().clearGlobalFocusOwner();
                            }
                            updateWindowMenu(true);
                        }

                        public void internalFrameActivated(InternalFrameEvent e) { activeWindowChanged(); }
                        public void internalFrameDeactivated(InternalFrameEvent e) { activeWindowChanged(); }
                        public void internalFrameDeiconified(InternalFrameEvent e) { activeWindowChanged(); }
                        public void internalFrameIconified(InternalFrameEvent e) { activeWindowChanged(); }
                        public void internalFrameOpened(InternalFrameEvent e) { }
                        public void internalFrameClosing(InternalFrameEvent e) { }
                    });

                    frame.setBounds(internalFrameCounter*30, internalFrameCounter*20, MINIMIZEDWIDTH, MINIMIZEDHEIGHT);
                    internalFrameCounter = (internalFrameCounter+1) % 10;

                    try {   // Minimise other frames
                        JInternalFrame[] fs = desktop.getAllFrames();
                        for (int i=0;i<fs.length;i++) {
                            fs[i].setMaximum(false);
                        }
                        desktop.add(frame);
                        frame.setVisible(true);
                        frame.setMaximum(fs.length==0);
                    } catch (PropertyVetoException e) { }
                    PDF pdf = parser.getPDF();
                    PDFPage initialPage = pdf!=null && pagenumber >=0 && pagenumber < pdf.getNumberOfPages() ? pdf.getPage(pagenumber) : null;
                    documentpanel.setPDF(parser, initialPage);
                    updateWindowMenu(true);
                } else {
                    if (createdEventPending) {
                        for (Iterator i = listeners.iterator();i.hasNext();) {
                            activedocument.addDocumentPanelListener((DocumentPanelListener)i.next());
                        }
                        activedocument.raiseDocumentPanelEvent(DocumentPanelEvent.createCreated(activedocument));
                        createdEventPending = false;
                    }
                    activedocument.putClientProperty("file", file);
                    PDF pdf = parser.getPDF();
                    PDFPage initialPage = pdf!=null && pagenumber >=0 && pagenumber < pdf.getNumberOfPages() ? pdf.getPage(pagenumber) : null;
                    activedocument.setPDF(parser, initialPage);
                    for (Iterator i = pdfcomponents.iterator();i.hasNext();) {
                        JComponent comp = (JComponent)i.next();
                        comp.setEnabled(true);
                    }
                }
                if (file!=null) {
                    ViewerFeature feature = getFeature(OpenRecent.class);
                    if (feature!=null) {
                        try {
                            ((OpenRecent)feature).opened(PDFViewer.this, file.getCanonicalFile());
                        } catch (IOException e) {}
                    }
                }
            }
        };
        if (SwingUtilities.isEventDispatchThread()) {
            r.run();
        } else {
            SwingUtilities.invokeLater(r);
        }
    }

    /**
     * Update the "Window" menu with the list of currently displayed documents
     * @param rebuild whether a rebuild is required due to a window being opened or closed
     */
    private void updateWindowMenu(boolean rebuild) {
        if (hasFeature(Menus.getInstance()) && hasFeature(ActiveWindowMenu.getInstance())) {
            JMenu menu = getMenu("Window");
            if (menu != null) {
                if (rebuild) {
                    for (int i=0;i<menu.getMenuComponentCount();i++) {
                        if (menu.getMenuComponent(i) instanceof JPopupMenu.Separator) {
                            while (i<menu.getMenuComponentCount()) {
                                menu.remove(i);
                            }
                        }
                    }
                    menu.addSeparator();
                    DocumentPanel[] panels = getDocumentPanels();
                    for (int i=0;i<panels.length;i++) {
                        final JInternalFrame frame = (JInternalFrame)SwingUtilities.getAncestorOfClass(JInternalFrame.class, panels[i]);
                        if (frame != null) {
                            String name = frame.getTitle();
                            JMenuItem item = new JCheckBoxMenuItem(name, false);
                            item.addActionListener(new ActionListener() {
                                public void actionPerformed(ActionEvent event) {
                                    try {
                                        frame.setSelected(true);
                                    } catch (PropertyVetoException e) {}
                                }
                            });
                            menu.add(item);
                        }
                    }
                }
                JInternalFrame active = desktop.getSelectedFrame();
                for (int i=0;i<menu.getMenuComponentCount();i++) {
                    if (menu.getMenuComponent(i) instanceof JCheckBoxMenuItem) {
                        JCheckBoxMenuItem item = (JCheckBoxMenuItem)menu.getMenuComponent(i);
                        item.setSelected(active!=null && item.getText().equals(active.getTitle()));
                    }
                }
            }
        }
    }

    /**
     * Return the active {@link DocumentPanel}
     */
    public DocumentPanel getActiveDocumentPanel() {
        initialize();
        return activedocument;
    }

    /**
     * Return all the {@link DocumentPanel}s in the viewer
     * @since 2.8.5
     */
    public DocumentPanel[] getDocumentPanels() {
        if (hasFeature(MultiWindow.getInstance())) {
            if (desktop == null) {
                return null;
            } else {
                JInternalFrame[] fs = desktop.getAllFrames();
                DocumentPanel[] panels = new DocumentPanel[fs.length];
                for (int i=0;i<fs.length;i++) {
                    try {
                        panels[i] = (DocumentPanel)(fs[i].getContentPane());
                    } catch (ClassCastException e) {
                        DocumentPanel[] t = new DocumentPanel[panels.length-1];
                        System.arraycopy(panels, 0, t, 0, i);
                        panels = t;
                    }
                }
                return panels;
            }
        } else {
            return new DocumentPanel[] { activedocument };
        }
    }

    /**
     * Close the PDFViewer. Should be called before this PDFViewer
     * is permanently removed from the Swing object hierarchy.
     * @since 2.11.18
     */
    public void close() {
        if (hasFeature(MultiWindow.getInstance())) {
            JInternalFrame[] fs = desktop.getAllFrames();
            for (int i=0;i<fs.length;i++) {
                fs[i].doDefaultCloseAction();
            }
        } else if (activedocument != null) {
            activedocument.setPDF((PDFParser)null, null);
        }
        for (Iterator i = features.iterator();i.hasNext();) {
            ViewerFeature feature = (ViewerFeature)i.next();
            feature.teardown();
        }
        if (menubar != null) {
            menubar.removeAll();
        }
    }

    /**
     * Close the specified {@link DocumentPanel}
     * @param panel the panel to close - usually the return value of {@link #getActiveDocumentPanel}
     * @since 2.10.2
     */
    public void closeDocumentPanel(DocumentPanel panel) {
        if (hasFeature(MultiWindow.getInstance())) {
            if (panel != null) {
                JInternalFrame window = (JInternalFrame)SwingUtilities.getAncestorOfClass(JInternalFrame.class, panel);
                if (window != null) {
                    window.doDefaultCloseAction();
                }
            }
        } else if (panel == activedocument) {
            if (!getUnpromptedDirtyClose() && panel.isDirty() && JOptionPane.showConfirmDialog(this, UIManager.getString("PDFViewer.ConfirmCloseText"), UIManager.getString("PDFViewer.Confirm"), JOptionPane.YES_NO_OPTION) == JOptionPane.NO_OPTION) {
                return;
            }

            activedocument.setPDF(null);
            activedocument.raiseDocumentPanelEvent(DocumentPanelEvent.createDeactivated(activedocument));
            for (Iterator i = pdfcomponents.iterator();i.hasNext();) {
                JComponent comp = (JComponent)i.next();
                comp.setEnabled(false);
            }
        }
    }

    /**
     * Add a {@link DocumentPanelListener} to any {@link DocumentPanel} objects
     * created by this PDFViewer.
     */
    public void addDocumentPanelListener(DocumentPanelListener listener) {
        initialize();
        listeners.add(listener);
        if (desktop!=null) {
            JInternalFrame[] fs = desktop.getAllFrames();
            for (int i=0;i<fs.length;i++) {
                try {
                    DocumentPanel panel = (DocumentPanel)(fs[i].getContentPane());
                    panel.addDocumentPanelListener(listener);
                } catch (ClassCastException e) {}
            }
        } else if (activedocument!=null && createdEventPending) {
            activedocument.addDocumentPanelListener(listener);
        }
    }

    /**
     * Add a {@link DocumentPanelListener} to any {@link DocumentPanel} objects
     * created by this PDFViewer.
     */
    public void removeDocumentPanelListener(DocumentPanelListener listener) {
        initialize();
        listeners.remove(listener);
        if (desktop!=null) {
            JInternalFrame[] fs = desktop.getAllFrames();
            for (int i=0;i<fs.length;i++) {
                try {
                    DocumentPanel panel = (DocumentPanel)(fs[i].getContentPane());
                    panel.removeDocumentPanelListener(listener);
                } catch (ClassCastException e) {}
            }
        } else if (activedocument!=null) {
            activedocument.removeDocumentPanelListener(listener);
        }
    }

    //----------------------------------------------------------------------------------

    private void cascadeFrames() {
        JInternalFrame[] frames = desktop.getAllFrames();
        if (frames.length != 0) {
            Rectangle dBounds = desktop.getBounds();
            int margin = 0;
            Insets ins = null;
            Dimension d, cd;
            for (int i=0; i<frames.length; i++) {
                d = frames[i].getSize();
                cd = frames[i].getContentPane().getSize();
                margin += Math.max(d.width - cd.width, d.height - cd.height);
            }

            int width = dBounds.width - margin;
            int height = dBounds.height - margin;
            int offset = 0;
            for (int i=0;i<frames.length;i++) {
                try {
                    frames[i].setMaximum(false);
                } catch (PropertyVetoException e) {}
                frames[i].setBounds(dBounds.x + offset, dBounds.y + offset, width, height);
                d = frames[i].getSize();
                cd = frames[i].getContentPane().getSize();
                offset += Math.max(d.width - cd.width, d.height - cd.height);
            }
        }
    }

    //----------------------------------------------------------------------------------

    /**
     * Create a new PDFViewer object in a frame of it's own. Top level routine
     * to be called by main()
     * @since 2.7..1
     */
    public static PDFViewer newPDFViewer() {
        ArrayList f = new ArrayList(ViewerFeature.getAllEnabledFeatures());
//        f.remove(MultiWindow.getInstance());
        return newPDFViewer(f);
    }

    /**
     * Create a new PDFViewer object in a frame of it's own with the specified features.
     * @param features a Collection of {@link ViewerFeature} objects that are to be supported
     * @see ViewerFeature
     * @since 2.7.9
     */
    public static PDFViewer newPDFViewer(Collection features) {
        JFrame frame = new JFrame("BFO");
        frame.setIconImage(BFO16.getImage());
        PDFViewer viewer = new PDFViewer(features);
        frame.addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) {
                System.exit(0);
            }
        });

        frame.setContentPane(viewer);
        frame.pack();
        Dimension screen = viewer.getToolkit().getScreenSize();
        Dimension window = frame.getSize();
        frame.setLocation(Math.max(0, (screen.width-window.width) / 2), Math.max(0, (screen.height-window.height) / 2));
        frame.applyComponentOrientation(ComponentOrientation.getOrientation(frame.getLocale()));
        frame.setVisible(true);
        return viewer;
    }

    /**
     * The main() method can be invoked to run this class from the command line.
     * A single argument specifying the name of the file to open is optional
     */
    public static void main(final String args[]) {
        File file = null;
        int pagenumber = 0;
        if (args.length > 0) {
            try {
                if (args.length > 3 || args[0].equals("-help")) {
                    System.err.println("Usage: java org.faceless.pdf2.viewer2.PDFViewer [<file>]");
                    return;
                } else {
                    // load PDF
                    if (args.length == 1 || args.length == 3) {
                        String filename = args[args.length-1];
                        int ix = filename.indexOf("#");
                        if (ix > 0) {
                            try {
                                pagenumber = Integer.parseInt(filename.substring(ix+1)) - 1;
                                filename = filename.substring(0, ix);
                            } catch (Exception e) { }
                        }
                        file = new File(filename);

                        if (!file.isFile() || !file.canRead()) {
                             System.err.println("PDFViewer: Unable to read \""+filename+"\"");
                             file = null;
                        }
                    }

                    if (args[0].equals("-license")) {
                        PDF.setLicenseKey(args[1]);
                    }
                }
            } catch (Exception e) {
                System.err.println("Usage: java org.faceless.pdf2.viewer2.PDFViewer [<file>]");
                return;
            }
        }

        try {
            if (System.getProperty("swing.defaultlaf") == null) {
                UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
            }
        } catch (Exception e) { }


        final File ffile = file;
        final int fpagenumber = pagenumber;
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                PDFViewer viewer = newPDFViewer();
                try {
                    KeyStoreTrustManager.install(viewer);
                } catch (Exception e) {
                    e.printStackTrace();
                }
                if (ffile != null) {
                    viewer.loadPDF(ffile, fpagenumber);
                }
            }
        });
    }

    /**
     * Set the name of the current user (for Annotations). The
     * default is null.
     * Can also be set by the <code>currentUser</code> initialization parameter
     * @since 2.11.18
     */
    public void setCurrentUser(String user) {
        this.user = user;
    }

    /**
     * Get the name of the current user, as set by {@link #setCurrentUser}
     * @since 2.11.18
     */
    public String getCurrentUser() {
        return this.user;
    }

    /**
     * Set whether to allow a window containing a modified PDF to be closed
     * without prompting.
     * Can also be set by the <code>unpromptedDirtyClose</code> initialization parameter
     * @since 2.11.19
     */
    public void setUnpromptedDirtyClose(boolean ignore) {
        this.unprompteddirtyclose = ignore;
    }

    /**
     * Return the value set by {@link #setUnpromptedDirtyClose}
     * @since 2.11.19
     */
    public boolean getUnpromptedDirtyClose() {
        return unprompteddirtyclose;
    }

}
