// $Id: PagePanel.java 15987 2012-07-03 13:26:41Z mike $

package org.faceless.pdf2.viewer2;

import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.*;
import java.util.*;
import java.util.List;
import javax.swing.*;
import org.faceless.util.SoftInterruptibleThread;
import org.faceless.pdf2.*;

/**
 * <p>
 * The <code>PagePanel</code> class is the lowest-level class for rendering a {@link PDFPage}
 * as a {@link JPanel}. At it's most basic it will simply render the page via a {@link PagePainter},
 * but when included inside a {@link DocumentViewport} as part of a {@link DocumentPanel}
 * this class may also create subcomponents representing {@link PDFAnnotation}s, as created by the
 * {@link AnnotationComponentFactory} class.
 * </p><p>
 * Before a PagePanel is disposed of it should have <code>setParser(null)</code> called on it.
 * </p>
 * See the <a href="doc-files/tutorial.html">viewer tutorial</a> for more detail on how to use this class and the "viewer" package.
 * <p><i>This code is copyright the Big Faceless Organization. You're welcome to use, modify and distribute it in any form in your own projects, provided those projects continue to make use of the Big Faceless PDF library.</i></p>
 * @see DocumentViewport
 * @since 2.8
 */
public class PagePanel extends JPanel {

    private static final long DEFAULTMAXANNOTATIONPIXELS = 1700 * 2800;
    private long MAXANNOTATIONPIXELS = DEFAULTMAXANNOTATIONPIXELS;        // Max size of annotation bitmap
    private static final long MIN_PIXELS_FOR_CLIP = 2097152L; // Enough for 100% A4 at 150dpi

    private DocumentViewport viewport;          // The viewport this page is a member of
    private PDFPage page;                       // The page currently being displayed
    private PDFParser parser;                   // The PDFParser to create the pages
    private PagePainter pagepainter;            // The PagePainter to paint the current page
    private PainterThread paintthread;          // Thread that handles the page update
    private volatile BufferedImage pageimage;            // The image of the page created in paintthread
    private volatile Rectangle pageclip;

    private volatile float dpi;                          // The DPI
    private volatile float x1, y1, x2, y2;               // Scope of the viewport
    private volatile int orientation;                    // The Page orientation of the last paint

    private RenderingHints hints;
    private final Collection listeners;         // Collection of PagePanelListener objects
    private final Collection ilisteners;        // Collection of PagePanelInteractionListener objects
    private Map annotcomponents;                // Map of PDFAnnotation -> JComponent.  Modify only in EDT!
    private Listener mouselistener;
    private FormListener formlistener;
    private boolean extract;
    private AffineTransform pageToScreen, screenToPage; // Transforms between page<=>screen coordinates

    private final int panelindex;               // For debugging
    private static int globalpanelindex;

    /**
     * Create a new PagePanel
     */
    public PagePanel() {
        super(null, true);
        Util.initialize();
        setOpaque(false);
        setFocusTraversalPolicy(this.new AnnotationFocusTraversalPolicy());
        listeners = new LinkedHashSet();
        ilisteners = new LinkedHashSet();
        annotcomponents = Collections.synchronizedMap(new HashMap());
        mouselistener = new Listener();
        formlistener = new FormListener();
        panelindex = globalpanelindex++;
        setFocusable(false);

        // Click the panel to remove focus from an editable annotation
        // widget
        addMouseListener(new MouseAdapter() {
            public void mousePressed(MouseEvent event) {
                JComponent viewport = (JComponent)SwingUtilities.getAncestorOfClass(DocumentViewport.class, PagePanel.this);
                if (viewport != null) {
                    viewport.requestFocusInWindow();
                }
            }
        });
    }

    /**
     * Return a read-only mapping from PDFAnnotation->JComponent
     */
    Map getAnnotationComponents() {
        return Collections.unmodifiableMap(annotcomponents);
    }

    /**
     * Return the area of the page that is considered to be the "whole page"
     * as far as the viewer is concerned - the {@link PDFPage#getBox ViewBox}
     * @return the area of the page that's the full page, in points
     * @see PDFPage#getBox
     * @see PDF#getOption
     */
    public static Rectangle2D getFullPageView(PDFPage page) {
        float[] box = page.getBox("ViewBox");
        return new Rectangle2D.Float(box[0], box[1], box[2]-box[0], box[3]-box[1]);
    }

    /**
     * Add a {@link PagePanelListener} to this PagePanel
     */
    public void addPagePanelListener(PagePanelListener listener) {
        if (listener!=null) {
            synchronized(listeners) {
                listeners.add(listener);
            }
        }
    }

    /**
     * Remove a {@link PagePanelListener} from this PagePanel
     */
    public void removePagePanelListener(PagePanelListener listener) {
        synchronized(listeners) {
            listeners.remove(listener);
        }
    }

    /**
     * Add a {@link PagePanelInteractionListener} to this PagePanel
     */
    public void addPagePanelInteractionListener(PagePanelInteractionListener listener) {
        if (listener!=null) {
            synchronized(ilisteners) {
                if (ilisteners.isEmpty()) {
                    addMouseListener(mouselistener);
                    addMouseMotionListener(mouselistener);
                }
                ilisteners.add(listener);
            }
        }
    }

    /**
     * Remove a {@link PagePanelInteractionListener} from this PagePanel
     */
    public void removePagePanelInteractionListener(PagePanelInteractionListener listener) {
        synchronized(ilisteners) {
            ilisteners.remove(listener);
            if (ilisteners.isEmpty()) {
                removeMouseListener(mouselistener);
                removeMouseMotionListener(mouselistener);
            }
        }
    }

    /**
     * Set the {@link RenderingHints} to be used when rendering pages
     * in this PagePanel.
     * @param hints the RenderingHints to use - may be null
     */
    public void setRenderingHints(RenderingHints hints) {
        this.hints = hints;
    }

    /**
     * Set the {@link PDFParser} object which should be used to render the pages.
     * Setting this to null will end the background thread that renders this object,
     * and free any related resources.
     * @param parser the PDFParser
     */
    public synchronized void setParser(PDFParser parser) {
        if (parser==null) {
            if (paintthread!=null) {
                paintthread.cancel();
            }
            this.page = null;
            this.pagepainter = null;
        } else {
            if (paintthread==null) {
                paintthread = new PainterThread();
                paintthread.start();
            }
        }
        this.parser = parser;
    }

    /**
     * Set whether to extract text while rendering this page to a {@link PageExtractor}
     * @param extract true to extract the text, false otherwise
     */
    public void setExtractText(boolean extract) {
        this.extract = extract;
    }

    /**
     * Return the {@link PageExtractor} associated with this PagePanel.
     * To render and extract text in a single pass, the {@link #setExtractText}
     * method should be called before rendering, otherwise this method
     * will run the extraction process in a separate pass.
     */
    public synchronized PageExtractor getPageExtractor() {
        return parser == null || page == null ? null : parser.getPageExtractor(page);
    }

    /**
     * Redraw the page. The page will be rerendered in the background and when it's
     * complete this component will be resized and repainted. If annotations are being
     * created they will be repositioned and redrawn as well.
     * This method may be called from any thread.
     *
     * @param page the page to draw
     * @param position the area of the page to draw, in points
     * @param dpi the resolution
     */
    public void setPage(PDFPage page, Rectangle2D position, float dpi) {
        setPage(page, (float)position.getMinX(), (float)position.getMinY(), (float)position.getMaxX(), (float)position.getMaxY(), dpi);
    }

    /**
     * Redraw the page. The page will be rerendered in the background and when it's
     * complete this component will be resized and repainted. If annotations are being
     * created they will be repositioned and redrawn as well.
     * This method may be called from any thread.
     *
     * @param page the page to draw
     * @param position the area of the page to draw, in points
     * @param dpi the resolution
     * @param clip The graphics clip for this panel, in pixels. The panel will not
     * not render any of the page outside this rectangle.
     * @since 2.10.4
     */
    public void setPage(PDFPage page, Rectangle2D position, float dpi, Rectangle clip) {
        setPage(page, (float)position.getMinX(), (float)position.getMinY(), (float)position.getMaxX(), (float)position.getMaxY(), dpi, clip);
    }

    /**
     * Redraw the page. The page will be rerendered in the background and when it's
     * complete this component will be resized and repainted. If annotations are being
     * created they will be repositioned and redrawn as well.
     * This method may be called from any thread.
     *
     * @param page the page to draw
     * @param x1 the left-most X co-ordinate to draw, in points from the bottom-left
     * @param y1 the bottom-most Y co-ordinate to draw, in points from the bottom-left
     * @param x2 the right-most X co-ordinate to draw, in points from the bottom-left
     * @param y2 the top-most Y co-ordinate to draw, in points from the bottom-left
     * @param dpi the resolution
     */
    public synchronized void setPage(PDFPage page, float x1, float y1, float x2, float y2, float dpi) {
        setPage(page, x1, y1, x2, y2, dpi, null);
    }

    /**
     * Redraw the page. The page will be rerendered in the background and when it's
     * complete this component will be resized and repainted. If annotations are being
     * created they will be repositioned and redrawn as well.
     * This method may be called from any thread.
     *
     * @param page the page to draw
     * @param x1 the left-most X co-ordinate to draw, in points from the bottom-left
     * @param y1 the bottom-most Y co-ordinate to draw, in points from the bottom-left
     * @param x2 the right-most X co-ordinate to draw, in points from the bottom-left
     * @param y2 the top-most Y co-ordinate to draw, in points from the bottom-left
     * @param dpi the resolution
     * @param clip the graphics clip for this panel in pixels. The panel will not
     * render any of the page outside this rectangle.
     * @since 2.10.4
     */
    public synchronized void setPage(PDFPage page, float x1, float y1, float x2, float y2, float dpi, Rectangle clip) {
        if (parser==null) {
            throw new IllegalStateException("No Parser set");
        }
        if (dpi<=0 || dpi!=dpi) {
            throw new IllegalArgumentException("Invalid DPI "+dpi);
        }
        if (page != this.page) {
            if (page != null) {
                page.flush();
                page.addPropertyChangeListener(formlistener);
            }
            if (this.page != null) {
                this.page.removePropertyChangeListener(formlistener);
            }
        }
        paintthread.redraw(page, x1, y1, x2, y2, dpi, clip, false);
    }

    /**
     * Force a redraw of the current page
     * @since 2.11.7
     */
    public void redrawCurrentPage() {
        if (paintthread != null) {
            paintthread.redrawCurrentPage();
        }
    }

    /**
     * Get the {@link DocumentViewport} this PagePanel is contained inside, or
     * <code>null</code> if this PagePanel was not created as part of a
     * {@link DocumentPanel}
     */
    public DocumentViewport getViewport() {
        return viewport;
    }

    /**
     * Get the {@link DocumentPanel} this PagePanel is contained inside, or
     * <code>null</code> if this PagePanel was not created as part of a
     * {@link DocumentPanel}. A shortcut for <code>getViewport().getDocumentPanel()</code>
     */
    public DocumentPanel getDocumentPanel() {
        return viewport==null ? null : viewport.getDocumentPanel();
    }

    /**
     * Return the {@link PDFPage} currently being displayed by this viewport.
     * Note this method returns the page being displayed, not the page currently
     * being rendered, so it's value will not immediately reflect the page passed
     * in to {@link #setPage setPage()}, and will be <code>null</code> if the
     * first page has not yet finished rendering.
     */
    public synchronized PDFPage getPage() {
        return page;
    }

    /**
     * Return the clip area actually used by the PagePanel. This may
     * be what was requested by the last call to {@link #setPage setPage},
     * or it may be <code>null</code> if the PagePanel elected to draw
     * the entire page instead.
     * @since 2.11.25
     */
    public synchronized Rectangle2D getClip() {
        return pageclip;
    }

    /**
     * Returns the area of the page currently being displayed, in points with
     * (0,0) at the bottom-left. Like {@link #getPage} the return value of
     * this method will not immediately reflect the position passed in to
     * {@link #setPage setPage()}, and will be <code>null</code> if the first
     * page has not yet finished rendering.
     */
    public synchronized Rectangle2D getView() {
        Rectangle2D rect = null;
        if (page!=null) {
            int rot = (page.getPageOrientation()-orientation+360)%360;
            if (rot==0) {
                rect = new Rectangle2D.Float(x1, y1, x2-x1, y2-y1);
            } else {
                int w = page.getWidth();
                int h = page.getHeight();
                if (rot==90) {
                    rect = new Rectangle2D.Float(y1, h-x2, y2-y1, x2-x1);
                } else if (rot==180) {
                    rect = new Rectangle2D.Float(w-x2, h-y2, x2-x1, y2-y1);
                } else {
                    rect = new Rectangle2D.Float(h-y2, x1, y2-y1, x2-x1);
                }
            }
        }
        return rect;
    }

    /**
     * Return an AffineTransform that will map Page co-ordinates to Screen
     * co-ordinates.
     * @since 2.10.2
     */
    public synchronized AffineTransform getPageToScreenTransform() {
        return pageToScreen;
    }

    /**
     * Return an AffineTransform that will map Page co-ordinates to Screen
     * co-ordinates.
     * @since 2.10.3
     */
    public synchronized AffineTransform getScreenToPageTransform() {
        return screenToPage;
    }


    static boolean equalsRectangle2D(Rectangle2D r1, Rectangle2D r2) {  // To catch rounding error
        double x1d = Math.abs(r1.getMinX()-r2.getMinX());
        double x2d = Math.abs(r1.getMaxX()-r2.getMaxX());
        double y1d = Math.abs(r1.getMinY()-r2.getMinY());
        double y2d = Math.abs(r1.getMaxY()-r2.getMaxY());
        return x1d<0.01 && x2d<0.01 && y1d<0.01 && y2d<0.01;
    }

    /**
     * Return the resolution of the page currently being displayed.
     * Like {@link #getPage} the return value of this method will not
     * immediately reflect the position passed in to {@link #setPage setPage()},
     * and will be 0 if the fist page has not yet finished rendering.
     */
    public synchronized float getDPI() {
        return page==null ? 0 : dpi;
    }

    /**
     * Given a location on this panel in pixels, return the equivalent
     * position on the current page in points.
     * @see #getAWTPoint
     */
    public synchronized Point2D getPDFPoint(int x, int y) {
        AffineTransform t = getScreenToPageTransform();
        Point2D.Float f = new Point2D.Float(x, y);
        return t == null ? null : t.transform(f, f);
    }

    /**
     * Given a location on this panel in pixels, return the equivalent
     * position on the current page in points.
     * @see #getAWTPoint
     */
    public synchronized Point2D getPDFPoint(Point awt) {
        AffineTransform t = getScreenToPageTransform();
        Point2D.Float f = new Point2D.Float(awt.x, awt.y);
        return t == null ? null : t.transform(f, f);
    }

    /**
     * Given a location on the page in points, return the equivalent
     * position on this PagePanel in pixels.
     * @see #getPDFPoint
     */
    public synchronized Point getAWTPoint(float x, float y) {
        AffineTransform t = getPageToScreenTransform();
        Point2D.Float f = new Point2D.Float(x, y);
        if (t == null) return null;
        t.transform(f, f);
        return new Point((int)Math.round(f.x), (int)Math.round(f.y));
    }

    /**
     * Given a location on the page in points, return the equivalent
     * position on this PagePanel in pixels.
     * @see #getPDFPoint
     */
    public synchronized Point getAWTPoint(Point2D f) {
        AffineTransform t = getPageToScreenTransform();
        f = (Point2D)f.clone();
        if (t == null) return null;
        t.transform(f, f);
        return new Point((int)Math.round(f.getX()), (int)Math.round(f.getY()));
    }

    public Dimension getPreferredSize() {
        if (pageimage == null) {
            return super.getPreferredSize();
        } else {
            return new Dimension(pageimage.getWidth(), pageimage.getHeight());
        }
    }

    /**
     * Set the viewport this PagePanel belongs to
     * @since 2.11
     */
    public void setViewport(DocumentViewport viewport) {
        this.viewport = viewport;
        if (viewport != null) {
            PDFViewer viewer = viewport.getDocumentPanel().getViewer();
             PropertyManager manager = viewer == null ? PDF.getPropertyManager() : viewer.getPropertyManager();
             try {
                 MAXANNOTATIONPIXELS = Long.parseLong(manager.getProperty("maxAnnotationBytes")) / 4;
             } catch (Exception e) {
                 MAXANNOTATIONPIXELS = DEFAULTMAXANNOTATIONPIXELS;
             }
        }
    }

    /**
     * After the repaint has completed, this method is called in
     * the EDT to update the values in this PagePainter object.
     */
    private synchronized void updatePagePosition(PDFPage page, PagePainter painter, float x1, float y1, float x2, float y2, float dpi, float zoom) {
        this.page = page;
        this.pagepainter = painter;
        this.x1 = x1;
        this.y1 = y1;
        this.x2 = x2;
        this.y2 = y2;
        this.dpi = dpi;
        this.orientation = page.getPageOrientation();
        pageToScreen = new AffineTransform(zoom, 0, 0, -zoom, -x1*zoom, y2*zoom);
        try {
            screenToPage = pageToScreen.createInverse();
        } catch (NoninvertibleTransformException e) {
            throw new Error();  // Can't happen
        }
    }

    private void raisePagePanelEvent(PagePanelEvent event) {
        boolean debug = (getDocumentPanel()==null || getDocumentPanel().getViewer()==null ? PDF.getPropertyManager() : getDocumentPanel().getViewer().getPropertyManager()).getProperty("debug.Event")!=null;
        if (debug) {
            if (SwingUtilities.isEventDispatchThread()) {
                System.err.println("Raise PagePanel#"+panelindex+" "+event);
            } else {
                System.err.println("Raise PagePanel#"+panelindex+" "+event+" (not in event thread)");
            }
        }
        PagePanelListener[] l = new PagePanelListener[0];
        synchronized(listeners) {
            l = (PagePanelListener[])listeners.toArray(l);
        }
        for (int i=0;i<l.length;i++) {
            l[i].pageUpdated(event);
        }
    }

    //------------------------------------------------------------------------------------
    // Methods dealing with annotations
    //------------------------------------------------------------------------------------

    protected void addImpl(Component c, Object constraints, int index) {
        super.addImpl(c, constraints, index);
        updateComponent((JComponent)c, null);
    }

    /**
     * For those JComponents based on the position of a PDFAnnotation
     * or a PDF-space rectangle, adjust the components position, visibility
     * and any other attributes that may be required.
     * Only called from the EDT.
     *
     * @param comp the JComponent, which will be adjusted if it has a
     * "pdf.annotation" or "pdf.rect" client property set.
     */
    private void updateComponent(JComponent comp, PDFPage page) {
        PDFAnnotation annot = (PDFAnnotation)comp.getClientProperty("pdf.annotation");
        float[] rect;
        if (annot!=null) {
            if (page!=null && page!=annot.getPage()) {
                comp.setVisible(false); // Needed, but don't know why.
                remove(comp);
                removeListeners(comp, annot);
                annotcomponents.remove(annot);
                rect = null;
            } else {
                comp.setVisible(annot.isVisible());
                rect = getAnnotationBounds(annot);
            }
        } else {
            rect = (float[])comp.getClientProperty("pdf.rect");
        }
        if (rect!=null) {
            boolean nozoom = annot instanceof AnnotationNote;
            float zoom = dpi/72f;
            float annotzoom = nozoom ? Math.min(zoom, 1) : zoom;
            int x = (int)Math.round((rect[0]-x1) * zoom);
            int y = (int)Math.round((y2-rect[3]) * zoom);
            int w = (int)Math.round((rect[2]-rect[0]) * annotzoom);
            int h = (int)Math.round((rect[3]-rect[1]) * annotzoom);
            if (w<0) { x+=w; w=-w; }
            if (h<0) { y+=h; h=-h; }
            comp.setBounds(x, y, w+1, h+1);
        }
    }

    /**
     * Remove any listeners from the annotation that refer to the Component.
     * Must be called before a component is freed, eg when the PagePanel is
     * discarded.
     */
    private void removeListeners(JComponent comp, PDFAnnotation annot) {
        annot.removePropertyChangeListener(formlistener);
        if (annot instanceof WidgetAnnotation) {
            FormElement field = ((WidgetAnnotation)annot).getField();
            field.removePropertyChangeListener(formlistener);
            PropertyChangeListener pcl = (PropertyChangeListener) comp.getClientProperty("bfo.fieldPCL");
            if (pcl != null) {
                field.removePropertyChangeListener(pcl);
            }
        }
    }

    /**
     * Return the bounds for this PDFAnnotation. The bounds is just
     * getRectangle(), unless it's an open AnnotationNote
     */
    private float[] getAnnotationBounds(PDFAnnotation annot) {
        float[] rect = annot.getRectangle();
        if (annot instanceof AnnotationNote) {
            rect = ((AnnotationNote)annot).getClosedRectangle();
        }
        return rect;
    }

    /**
     * Redraw the specified annotation on this page. This should be
     * called if the annotation is changed, if it's newly added to the
     * page or if it's been removed from the page.
     * May be called from any thread.
     *
     * @throws IllegalStateException if the PagePanel is not part of
     * a {@link DocumentPanel} and so has no annotations.
     */
    public void redrawAnnotation(final PDFAnnotation annot) {
        if (paintthread != null && annot != null) {
            paintthread.updateAnnotation(annot);
        }
    }

    /**
     * Return true if this PagePanel draws annotations as children
     */
    private boolean hasAnnotations() {
        return viewport!=null && viewport.getDocumentPanel()!=null && !viewport.getDocumentPanel().getAnnotationFactories().isEmpty();
    }

    /**
     * Create a JComponent for a PDFAnnotation. This calls the appropriate factory
     * for the annotation and then "binds" the location of the JComponent to the
     * PDFAnnotation by setting the "pdf.annotation" client property, which is
     * used by the methods in this class to link the JComponent with the PDFAnnotation
     * and position it correctly.
     * Only called from the EDT (as it creates components)
     */
    private JComponent createComponentForAnnotation(PDFAnnotation annot) {
        JComponent c = null;
        Collection factories = viewport.getDocumentPanel().getAnnotationFactories();
        for (Iterator i=factories.iterator();c==null && i.hasNext();) {
            AnnotationComponentFactory factory = (AnnotationComponentFactory)i.next();
            if (factory.matches(annot)) {
                c = factory.createComponent(this, annot);
            }
        }
        if (c==null) {
            c = AnnotationComponentFactory.getDefaultFactory().createComponent(this, annot);
        }
        AnnotationComponentFactory.bindComponentLocation(c, annot);
        annot.addPropertyChangeListener(formlistener);
        if (annot instanceof WidgetAnnotation) {
            FormElement field = ((WidgetAnnotation) annot).getField();
            field.addPropertyChangeListener(formlistener);
        }
        return c;
    }

    // This class handles dispatching events to the painter thread.
    private static class PainterQueue {

        private PainterEvent next;

        /**
         * Queue the next event. If the event is identical to the one we have on the
         * list already (same page, dpi and area) then we just append it's annotations
         * to the current list, otherwise we replace the current event queue with this one
         */
        synchronized void push(PainterEvent event) {
            if (next!=null && next.matches(event)) {
                if (next.repaintannotations==null) {
                    next.repaintannotations = event.repaintannotations;
                } else if (event.repaintannotations!=null) {
                    next.repaintannotations.addAll(event.repaintannotations);
                }
            } else {
                next = event;
            }
            notifyAll();
        }

        synchronized void cancel() {
            notifyAll();
        }

        synchronized PainterEvent pop() {
            if (isEmpty()) {
                try {
                    wait();
                } catch (InterruptedException e) { }
            }
            PainterEvent ret = next;
            next = null;
            return ret;
        }

        boolean isEmpty() {
            return next == null;
        }

    }

    private static class PainterEvent {

        float tdpi, tx1, ty1, tx2, ty2;         // The values we're going to redraw with
        boolean repaintcontent;
        List repaintannotations;

        PDFPage tpage;                          // These are promoted to the parent classes
        PagePainter tpainter;                   // permanent value when the draw is complete
        Rectangle clip;

        boolean matches(PainterEvent event) {
            return matches(event.tpage, event.tdpi, event.tx1, event.ty1, event.tx2, event.ty2);
        }

        boolean matches(PDFPage page, float dpi, float x1, float x2, float y1, float y2) {
            return page==tpage && dpi==tdpi && x1==tx1 && x2==tx2 && y1==ty1 && y2==ty2;
        }

        public String toString() {
            StringBuffer buf = new StringBuffer("PainterEvent[tdpi=");
            buf.append(tdpi);
            buf.append(",tx1=");
            buf.append(tx1);
            buf.append(",tx2=");
            buf.append(tx2);
            buf.append(",ty1=");
            buf.append(ty1);
            buf.append(",ty2=");
            buf.append(ty2);
            buf.append(",repaintcontent=");
            buf.append(repaintcontent);
            buf.append(",repaintannotations=");
            buf.append(repaintannotations==null?-1:repaintannotations.size());
            buf.append(",clip=");
            buf.append(clip);
            return buf.toString();
        }

    }

    //--------------------------------------------------------------------------------
    // The painter thread
    //--------------------------------------------------------------------------------
    private class PainterThread extends Thread implements SoftInterruptibleThread {

        private volatile boolean active;
        private PainterQueue queue;
        private PainterEvent lastEvent;
        private Rectangle lastClip;
        private volatile boolean interrupt; // whether to carry on rendering in this pass

        PainterThread() {
            super("PainterThread#"+panelindex);
            setDaemon(true);
            queue = new PainterQueue();
            active = true;
        }

        /**
         * If the paintthread is painting, interrupt it (if not this is a no-op)
         */
        private void interruptPaint() {
            interrupt = true;
        }

        public boolean isSoftInterrupted() {
            return !active || interrupt;
        } 

        /**
         * Force a redraw of the current page
         */
        synchronized void redrawCurrentPage() {
            if (lastEvent != null) {
                float[] box = lastEvent.tpage.getBox("ViewBox");
                redraw(lastEvent.tpage, box[0], box[1], box[2], box[3], lastEvent.tdpi, lastClip, true);
            }
        }

        /**
         * Kill the thread: interrupt what we're doing and signal thread to die after
         */
        private synchronized void cancel() {
            active = false;
            interrupt = true;
            if (lastEvent != null && lastEvent.tpainter!=null && lastEvent.tpainter.isPainting()) {
                lastEvent.tpainter.interrupt();
            }
            queue.cancel();
        }

        /**
         * Redraw with the specified page and position. One of three cases:
         * 1. If the values are identical, no action is taken
         * 2. If the page and DPI are identical, only page body is redrawn
         * 3. Otherwise page and all annotations are redrawn.
         *
         * For 2 and 3, if a paint is currently running it will be interrupted.
         * To initiate a redraw, we set (in a synchronized way) the redraw
         * parameters tx1, ty1, tpage etc., then notify the thread.
         */
        private synchronized void redraw(PDFPage page, float x1, float y1, float x2, float y2, float dpi, Rectangle clip, boolean forced) {
            PDFPage tpage = (lastEvent == null) ? null : lastEvent.tpage;
            boolean debug = (getDocumentPanel()==null || getDocumentPanel().getViewer()==null ? PDF.getPropertyManager() : getDocumentPanel().getViewer().getPropertyManager()).getProperty("debug.Event")!=null;

            // Decide whether to use the clip rectangle
            Dimension size = getSize();
            if (((long) size.width * (long) size.height) < MIN_PIXELS_FOR_CLIP) {
                // We may as well render the whole component as it is cheap
                // enough and we don't want to rerender with another
                // slightly different clip
                clip = null;
            }

            if (clip != null && lastClip != null) {
                Rectangle uclip = lastClip.union(clip);
                if (((long) uclip.width * (long) uclip.height) < MIN_PIXELS_FOR_CLIP) {
                    clip = uclip;
                }
            }

            if (!forced && lastEvent != null && lastEvent.matches(page, dpi, x1, x2, y1, y2) && (lastClip == clip || (clip != null && clip.equals(lastClip)))) {
                if (debug) {
                    System.out.println("PagePanel.redraw(P"+(page==null?-1:(page.getPageNumber()-1))+", lastEvent="+lastEvent+" duplicate");
                }
                return;
            }

            PainterEvent event = new PainterEvent();
            if ((page!=tpage || lastEvent==null || lastEvent.tdpi!=dpi) && page!=null && hasAnnotations()) {
                event.repaintannotations = new ArrayList(page.getAnnotations());
            }
            PagePainter painter = (lastEvent == null) ? null : lastEvent.tpainter;
            if (painter == null || page != tpage) {
                painter = parser.getPagePainter(page);
                if (hints != null) {
                    painter.setRenderingHints(hints);
                }
                if (hasAnnotations()) {
                    painter.setPaintAnnotations(false);
                }
            }
            event.tx1 = x1;
            event.tx2 = x2;
            event.ty1 = y1;
            event.ty2 = y2;
            event.tdpi = dpi;
            event.repaintcontent = true;
            event.tpage = page;
            event.tpainter = painter;
            event.clip = clip;

            if (debug) {
                System.out.println("PagePanel.redraw(P"+(page==null?-1:(page.getPageNumber()-1))+", event="+event);
            }
            interruptPaint();

            queue.push(event);
            lastClip = clip;
        }

        /**
         * Cause the painter to redraw the specified annotations
         */
        private synchronized void updateAnnotation(PDFAnnotation annot) {
            if (hasAnnotations()) {
                PDFPage page = (lastEvent == null) ? null : lastEvent.tpage;
                PagePainter painter = (lastEvent == null) ? null : lastEvent.tpainter;
                float dpi = (lastEvent == null) ? 0f : lastEvent.tdpi;

                if (page == null || painter == null || dpi == 0f) {
                    System.out.println("WARNING: no painter present while updating annotation");
                    return;
                }
                PainterEvent event = new PainterEvent();
                ArrayList annots = new ArrayList();
                annots.add(annot);
                event.repaintannotations = annots;
                event.tpage = page;
                event.tpainter = painter;
                event.tdpi = dpi;
                queue.push(event);
            } else {
                throw new IllegalStateException("PagePanel has no annotation children");
            }
        }

        /**
         * This method is the repaint thread body. It runs as a state machine, checking
         * "state": While it's not STOP we loop, waiting if it's WAIT or going ahead
         * immediately if it's INTERRUPTED. Once through we set to RUNNING and continue
         * so long as it remains that way, returning to WAIT if we get to the end.
         *
         * The thread first checks "repaintcontent". If true, it repaints the page body
         * using the values in this thread (tx1, ty1, tpage etc.).  If it completes without
         * interruption, it repaints (can do so immediately as this method is not contentious),
         * then stores those values temporarily and invokeLater() the next block in the EDT.
         *
         * Once we're in the EDT we can set the PagePainter values - they must be set here as
         * they are checked from many other places in the app, specifically the Viewports. Once
         * they are set we can inform everyone about it with the events. We also update the
         * annotation positions, even though they may not have their content rendered yet (if
         * we didn't render the pagecontent we just update the annotation positions).
         *
         * Meanwhile back in the render thread, once we've completed the page render (if needed)
         * we loop through the repaintannotations queue, getting the next one, rendering its
         * content and if we're not interrupted, removing it from the queue and repainting the
         * component for that annotation with the images we've generated. This may involve
         * creating a component if it doesn't already exist (which we do on the EDT again).
         *
         * A more efficient approach would be to create components for each annotation we're
         * going to render in one shot on the EDT...
         */
        public void run() {
            /* XXX test for "!isInterrupted()" when we use interrupt */
            while (active) {
//                System.out.println("PaintThread "+hashCode()+": waiting");
                PainterEvent event = (PainterEvent) queue.pop();
                interrupt = false;
                /* XXX test for "isInterrupted()" when we use interrupt */
                //if (isInterrupted())
                if (active && event != null) {
                    lastEvent = event;
//                    System.out.println("PaintThread "+hashCode()+": event="+event);

                    final PDFPage fpage = event.tpage;
                    final PagePainter fpainter = event.tpainter;
                    final float fx1 = event.tx1;
                    final float fx2 = event.tx2;
                    final float fy1 = event.ty1;
                    final float fy2 = event.ty2;
                    final float fdpi = event.tdpi;
                    final float zoom = event.tdpi / 72;

                    // Get the bounding rectangle of the clip in PDF space
                    float cfx1 = fx1, cfy1 = fy1, cfx2 = fx2, cfy2 = fy2;
                    Rectangle clip = event.clip;
                    if (clip != null) {
                        cfx1 = ((float) clip.x) / zoom;
                        cfy2 = (fy2 - fy1) - ((float) clip.y) / zoom;
                        cfx2 = cfx1 + ((float) clip.width) / zoom;
                        cfy1 = cfy2 - ((float) clip.height) / zoom;
                    }
                    try {
                        // Stage 1. Repaint the page content if required - once done, on the EDT
                        // reposition the annotations and raise the page events. Don't resize
                        // anymore, it's bad form to do it ourselves - we now return useful values
                        // from getPreferredSize() so we can do it in our parent.
                        //
                        if (event.repaintcontent && !interrupt && cfx2 > cfx1 && cfy2 > cfy1) {
                            SwingUtilities.invokeLater(new Runnable() {
                                public void run() {
                                    setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
                                }
                            });
                            BufferedImage img = null;
                            raisePagePanelEvent(PagePanelEvent.createPageRedrawing(PagePanel.this, fpage));
                            if (fpage==null) {
                                int iw = (int)Math.ceil((fx2-fx1)*zoom);
                                int ih = (int)Math.ceil((fy2-fy1)*zoom);
                                img = new BufferedImage(iw, ih, BufferedImage.TYPE_INT_ARGB);
                                Graphics g = img.getGraphics();
                                g.setColor(new Color(0xC0C0C0));
                                g.fillRect(0, 0, iw, ih);
                            } else if (fpainter != null) {
                                if (extract && parser != null) {
                                    fpainter.setPageExtractor(parser.getPageExtractor(fpage));
                                }
                                img = fpainter.getSubImage(cfx1, cfy1, cfx2, cfy2, fdpi, getColorModel());
                            }
                            if (img!=null && !interrupt) {            // Image was completed.
                                //System.out.println("PainterThread.run: setting pageimage for page "+fpage.getPageNumber());
                                pageimage = img;                                // First, repaint ASAP!
                                pageclip = clip;
                                repaint();
                                SwingUtilities.invokeLater(new Runnable() {     // Then, once in the EDT
                                    public void run() {                         // set some values:
                                        PDFPage oldpage = page;
                                        updatePagePosition(fpage, fpainter, fx1, fy1, fx2, fy2, fdpi, zoom);
                                        for (int i=getComponentCount()-1;active && i>=0;i--) {
                                            updateComponent((JComponent)getComponent(i), fpage);
                                        }
                                        // Now the values are all set, resize
                                        if (active && fpage!=oldpage) {       // Changed pages - remove all annots
                                            // Create Component here so they exist before the event
                                            // is fired. This is necessary so things that fire in
                                            // Page/Open event have annotations to work on.
                                            //
                                            // Note there's another EDT task trying to create these
                                            // annotations, below. This one will run first, but the
                                            // other task has already been scheduled by that point,
                                            // which is why we test twice for the Component already
                                            // existing in annotcomponents

                                            List annots = page.getAnnotations();
                                            for (int i=0;active && i<annots.size();i++) {
                                                PDFAnnotation annot = (PDFAnnotation)annots.get(i);
                                                JComponent comp = (JComponent)annotcomponents.get(annot);
                                                if (comp==null) {
                                                    comp = createComponentForAnnotation(annot);
                                                }
                                                if (comp!=null) {
                                                    add(comp);
                                                    annotcomponents.put(annot, comp);
                                                }
                                            }
                                            if (active) {
                                                repaint();
                                                raisePagePanelEvent(PagePanelEvent.createPageVisible(PagePanel.this, fpage));
                                                if (oldpage!=null) {
                                                    raisePagePanelEvent(PagePanelEvent.createPageHidden(PagePanel.this, oldpage));
                                                }
                                            }
                                        }
                                    }
                                });
                            }
                        } else if (!interrupt) {
                            SwingUtilities.invokeLater(new Runnable() {     // Update annot positions
                                public void run() {
                                    for (int i=getComponentCount()-1;active && i>=0;i--) {
                                        updateComponent((JComponent)getComponent(i), fpage);
                                    }
                                }
                            });
                        }

                        // Stage 2 - for any annotations in the "repaintannotations" list,
                        // draw the main "N" image, then attach to the component if there
                        // is one or create one on the next pass through the EDT if there isn't.
                        // Then set the "D" and "R" images if they exist.
                        //
                        // Only remove it from the list once we've succeeded drawing it, as
                        // we may be interrupted.
                        //
                        while (!interrupt && event.repaintannotations != null && !event.repaintannotations.isEmpty()) {
                            final PDFAnnotation annot = (PDFAnnotation) event.repaintannotations.remove(0);
                            final float[] rect = getAnnotationBounds(annot);
                            boolean nozoom = annot instanceof AnnotationNote;
                            float annotzoom = nozoom ? Math.min(1, zoom) : zoom;
                            int w = rect == null ? 0 : (int)Math.ceil((rect[2]-rect[0])*annotzoom)+1;
                            int h = rect == null ? 0 : (int)Math.ceil((rect[3]-rect[1])*annotzoom)+1;
                            if (((long)w * (long)h) > MAXANNOTATIONPIXELS && annot.getPage() == fpage) {
                                // The raster for the annotation is going to be too big to fit in memory.
                                // Use a smaller raster and scale the annotation onto it.
                                double scale = 1;
                                long w2, h2;
                                do {
                                    scale /= 2;
                                    w2 = (int)(scale * w);
                                    h2 = (int)(scale * h);
                                } while (w2 * h2 > MAXANNOTATIONPIXELS);
                                w = (int)w2;
                                h = (int)h2;
                                annotzoom *= scale;
                            }
                            if (w >= 1 && h >= 1 && annot.getPage() == fpage) {
                                AffineTransform tran = new AffineTransform(annotzoom, 0, 0, -annotzoom, -rect[0]*annotzoom, rect[3]*annotzoom);
                                BufferedImage image = new BufferedImage(w, h, BufferedImage.TYPE_INT_ARGB);
                                Graphics2D g = (Graphics2D)image.getGraphics();
                                if (hints!=null) g.setRenderingHints(hints);
                                g.setTransform(tran);
                                fpainter.setPageExtractor(null);

                                if (!interrupt && fpainter.paintAnnotation(annot, "N", g, rect)) {
                                    final BufferedImage[] images = new BufferedImage[] { image, null, null };
                                    final Composite[] composites = new Composite[] { fpainter.getAnnotationComposite(), null, null };

                                    JComponent comp = (JComponent)annotcomponents.get(annot);
                                    if (comp==null) {
                                        SwingUtilities.invokeLater(new Runnable() {
                                            public void run() {

                                                // Test the component exists again - this is because
                                                // on page load, the Component will have been created
                                                // already in this thread, just before the "redrawn"
                                                // event fired. We don't want to recreate it, but
                                                // readding it here to get the z-order is fine.
                                                //
                                                JComponent comp = (JComponent)annotcomponents.get(annot);
                                                if (active && comp==null) {
                                                    comp = createComponentForAnnotation(annot);
                                                }
                                                if (active && comp!=null && annot.getPage()!=null) {
                                                    annotcomponents.put(annot, comp);
                                                    int z = getAnnotationZOrder(annot);
                                                    add(comp, z==getComponentCount() ? -1 : z);
                                                    comp.putClientProperty("images", images);
                                                    comp.putClientProperty("imagecomposites", composites);
                                                }
                                            }
                                        });
                                    } else {
                                        final JComponent fcomp = comp;
                                        fcomp.putClientProperty("images", images);
                                        fcomp.putClientProperty("imagecomposites", composites);
                                        SwingUtilities.invokeLater(new Runnable() {
                                            public void run() {
                                                int z = getAnnotationZOrder(annot);
                                                try {
                                                    setComponentZOrder(fcomp, z);        // Java15
                                                } catch (Throwable e) {
                                                    add(fcomp, z==getComponentCount() ? -1 : z);
                                                }
                                                fcomp.repaint();
                                            }
                                        });
                                    }

                                    if (!interrupt && annot.hasAppearanceState("D")) {
                                        image = new BufferedImage(w, h, BufferedImage.TYPE_INT_ARGB);
                                        g = (Graphics2D)image.getGraphics();
                                        if (hints!=null) g.setRenderingHints(hints);
                                        g.setTransform(tran);
                                        if (!interrupt && fpainter.paintAnnotation(annot, "D", g, rect)) {
                                            images[1] = image;
                                            composites[1] = fpainter.getAnnotationComposite();
                                        }
                                    }

                                    if (!interrupt && annot.hasAppearanceState("R")) {
                                        image = new BufferedImage(w, h, BufferedImage.TYPE_INT_ARGB);
                                        g = (Graphics2D)image.getGraphics();
                                        if (hints!=null) g.setRenderingHints(hints);
                                        g.setTransform(tran);
                                        if (!interrupt && fpainter.paintAnnotation(annot, "R", g, rect)) {
                                            images[2] = image;
                                            composites[2] = fpainter.getAnnotationComposite();
                                        }
                                    }
                                }
                            } else {        // Too small! Remove the images
                                JComponent comp = (JComponent)annotcomponents.get(annot);
                                if (comp!=null) {
                                    comp.putClientProperty("images", null);
                                    comp.putClientProperty("imagecomposites", null);
                                    comp.repaint();
                                }
                            }
                        }
                        if (!interrupt) {
                            // We got the whole way through successfully. Raise the redrawn event
                            SwingUtilities.invokeLater(new Runnable() {
                                public void run() {
                                    if (isCursorSet() && getCursor() == Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR)) {
                                        setCursor(null);
                                    }
                                    raisePagePanelEvent(PagePanelEvent.createPageRedrawn(PagePanel.this, fpage));
                                }
                            });
                        }

                    } catch (RuntimeException e) {
                        Util.displayThrowable(e, PagePanel.this);
                    }
                }
            }
            paintthread = null;
            for (Iterator i = annotcomponents.entrySet().iterator();i.hasNext();) {
                Map.Entry e = (Map.Entry)i.next();
                PDFAnnotation annot = (PDFAnnotation)e.getKey();
                JComponent comp = (JComponent)e.getValue();
                removeListeners(comp, annot);
            }
            lastEvent = null;
        }

        private int getAnnotationZOrder(PDFAnnotation annot) {
            PDFPage page = annot.getPage();
            if (page!=null) {
                List annots = page.getAnnotations();
                int z = annots.size() - annots.indexOf(annot) - 1;
                return Math.max(0, Math.min(getComponentCount(), z));
            } else {
                return 0;
            }
        }
    }

    public void paintComponent(Graphics g) {
        super.paintComponent(g);
        g.setColor(getBackground());
        Rectangle bounds = getBounds();
        g.fillRect(bounds.x, bounds.y, bounds.width, bounds.height);
        if (pageimage != null) {
            int x = 0, y = 0, width = bounds.width, height = bounds.height;
            if (pageclip != null) {
                x = pageclip.x;
                y = pageclip.y;
                width = pageclip.width;
                height = pageclip.height;
            }
            g.drawImage(pageimage, x, y, width, height, this);
        }
    }

    private class Listener implements MouseMotionListener, MouseListener {
        private void raiseEvent(PagePanelInteractionEvent event) {
            PagePanelInteractionListener[] l = new PagePanelInteractionListener[0];
            synchronized(ilisteners) {
                l = (PagePanelInteractionListener[])ilisteners.toArray(l);
            }
            for (int i=0;i<l.length;i++) {
                l[i].pageAction(event);
            }
        }

        public void mouseEntered(MouseEvent event) {
            if (pageToScreen != null && screenToPage != null) {
                raiseEvent(PagePanelInteractionEvent.createMouseEntered(PagePanel.this, event));
            }
        }
        public void mouseExited(MouseEvent event) {
            if (pageToScreen != null && screenToPage != null) {
                raiseEvent(PagePanelInteractionEvent.createMouseExited(PagePanel.this, event));
            }
        }
        public void mousePressed(MouseEvent event) {
            if (pageToScreen != null && screenToPage != null) {
                raiseEvent(PagePanelInteractionEvent.createMousePressed(PagePanel.this, event));
            }
        }
        public void mouseReleased(MouseEvent event) {
            if (pageToScreen != null && screenToPage != null) {
                raiseEvent(PagePanelInteractionEvent.createMouseReleased(PagePanel.this, event));
            }
        }
        public void mouseClicked(MouseEvent event) {
            if (pageToScreen != null && screenToPage != null) {
                raiseEvent(PagePanelInteractionEvent.createMouseClicked(PagePanel.this, event));
            }
        }
        public void mouseMoved(MouseEvent event) {
            if (pageToScreen != null && screenToPage != null) {
                raiseEvent(PagePanelInteractionEvent.createMouseMoved(PagePanel.this, event));
            }
        }
        public void mouseDragged(MouseEvent event) {
            if (pageToScreen != null && screenToPage != null) {
                raiseEvent(PagePanelInteractionEvent.createMouseDragged(PagePanel.this, event));
            }
        }
    }

    private class FormListener implements PropertyChangeListener {

        public void propertyChange(PropertyChangeEvent event) {
            Object source = event.getSource();
            if (source == page) {
                String type = event.getPropertyName();
                if (type.equals("annotations")) {
                    PDFAnnotation oldannot = (PDFAnnotation)event.getOldValue();
                    PDFAnnotation newannot = (PDFAnnotation)event.getNewValue();
                    redrawAnnotation(oldannot);
                    redrawAnnotation(newannot);
                } else {
                    if (type.equals("orientation") || type.equals("content") || type.equals("mediabox")) {
                        parser.resetPageExtractor(page);
                    }
                    redrawCurrentPage();
                }
            } else if (source instanceof PDFAnnotation) {
                PDFAnnotation annot = (PDFAnnotation) source;
                redrawAnnotation(annot);
            } else if (source instanceof FormElement) {
                FormElement field = (FormElement) source;
                List annots = field.getAnnotations();
                for (Iterator i = annots.iterator(); i.hasNext(); ) {
                    PDFAnnotation annot = (PDFAnnotation) i.next();
                    redrawAnnotation(annot);
                }
            }
        }

    }

    private class AnnotationFocusTraversalPolicy extends FocusTraversalPolicy {

        private PDFAnnotation getAnnot(Component comp) {
            if (comp instanceof JComponent) {
                return (PDFAnnotation)((JComponent)comp).getClientProperty("pdf.annotation");
            } else {
                return null;
            }
        }

        public Component getComponentAfter(Container root, Component comp) {
            boolean seen = false;
            PDFAnnotation annot = getAnnot(comp);
            if (page != null) {
                List annots = page.getAnnotations();
                for (Iterator j = annots.iterator(); j.hasNext(); ) {
                    PDFAnnotation jannot = (PDFAnnotation) j.next();
                    if (seen) {
                        Component ret = (Component) annotcomponents.get(jannot);
                        return ret;
                    }
                    if (jannot == annot) {
                        seen = true;
                    }
                }
            }
            return root;
        }

        public Component getComponentBefore(Container root, Component comp) {
            boolean seen = false;
            PDFAnnotation annot = getAnnot(comp);
            if (page != null) {
                List annots = new ArrayList(page.getAnnotations());
                Collections.reverse(annots);
                for (Iterator j = annots.iterator(); j.hasNext(); ) {
                    PDFAnnotation jannot = (PDFAnnotation) j.next();
                    if (seen) {
                        Component ret = (Component) annotcomponents.get(jannot);
                        return ret;
                    }
                    if (jannot == annot) {
                        seen = true;
                    }
                }
            }
            return root;
        }

        public Component getDefaultComponent(Container root) {
            //return getFirstComponent(root);
            return root;
        }

        public Component getFirstComponent(Container root) {
            if (page != null) {
                List l = page.getAnnotations();
                if (!l.isEmpty()) {
                    Component ret = (Component)annotcomponents.get(l.get(0));
                    return ret;
                }
            }
            return root;
        }

        public Component getLastComponent(Container root) {
            Component lastComponent = root;
            if (page != null) {
                List l = page.getAnnotations();
                if (!l.isEmpty()) {
                    lastComponent = (Component)annotcomponents.get(l.get(l.size() - 1));
                }
            }
            return lastComponent;
        }

    }

}
