// $Id: ViewerWidget.java 15793 2012-06-11 13:30:29Z mike $

package org.faceless.pdf2.viewer2;

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.Container;
import java.util.*;
import java.beans.*;
import java.net.URL;
import org.faceless.pdf2.viewer2.feature.*;
import org.faceless.pdf2.PropertyManager;

/**
 * A type of ViewerFeature that adds a "widget" to a {@link PDFViewer}. Widgets are typically
 * buttons on the toolbar, menu items and so on.
 * <p><i>This code is copyright the Big Faceless Organization. You're welcome to use, modify and distribute it in any form in your own projects, provided those projects continue to make use of the Big Faceless PDF library.</i></p>
 * @see ToggleViewerWidget
 * @since 2.8
 */
public class ViewerWidget extends ViewerFeature
{
    private String toolbarname, icon, tooltip, menu;
    private char mnemonic;
    private boolean documentrequired, toolbarenabled, toolbarenabledalways, toolbarfloatable, floating;
    private JComponent component;
    private PDFViewer viewer;

    /**
     * Create a new Widget
     */
    public ViewerWidget(String name) {
        super(name);
        setDocumentRequired(true);
        setToolBarEnabled(true);
        setToolBarFloatable(true);
    }

    public String toString() {
        return "Widget:"+super.toString();
    }

    /**
     * Return an ActionListener that will be called when this Widget is activated.
     * Subclasses will typically not need to override this method except in special cases.
     * @see Quit#createActionListener
     * @return ActionListener the ActionListener to be notified when an event fires
     */
    protected ActionListener createActionListener() {
        return new ActionListener() {
            public void actionPerformed(ActionEvent evt) {
                action(new ViewerEvent(evt, viewer));
            }
        };
    }

    public void initialize(final PDFViewer viewer) {
        if (getViewer()!=null) {
            throw new IllegalStateException("Feature already added to another viewer");
        }
        this.viewer = viewer;

        ActionListener listener = createActionListener();

        PropertyManager propertymanager = viewer.getPropertyManager();
        String property = getClass().getName();
        if (property.startsWith("org.faceless.pdf2.viewer2.")) property = property.substring(26);

        if (toolbarname!=null) {
            boolean ltrsubpanel = false;
            if (toolbarname.endsWith(".ltr")) {
                // Hack to ensure page first/prev/number/right/last are in LTR order - we
                // put them into a subpanel which forces its ComponentOrientation
                toolbarname = toolbarname.substring(0, toolbarname.length() - 4);
                ltrsubpanel = true;
            }
            JComponent toolbar = viewer.getToolBar(toolbarname, toolbarenabled, toolbarenabledalways, toolbarfloatable, floating);
            String compname = getName();
            if (component == null) {
                JButton button;

                if (icon == null) {
                    throw new NullPointerException("No icon specified for button");
                }
                URL url;
                if (icon.length() > 0 && icon.charAt(0) == '/') {
                    url = getClass().getClassLoader().getResource(icon.substring(1));
                } else {
                    // Scan through each superclass for relative path icon: makes specifying
                    // these a little easier.
                    Class clazz = getClass();
                    do {
                        url = clazz.getResource(icon);
                        clazz = clazz.getSuperclass();
                    } while (url == null && clazz != ViewerFeature.class);
                }
                if (url == null) {
                    throw new NullPointerException("Can't find icon \""+icon+"\" for button");
                }
                button = new JButton(new ImageIcon(url)) {
                    public void paintComponent(Graphics g) {
                        super.paintComponent(g);
                        if (isSelected()) {
                            BorderFactory.createLoweredBevelBorder().paintBorder(this, g, 0, 0, getWidth()-1, getHeight()-1);
                            g.setColor(new Color(0x20000000, true));
                            g.fillRect(0, 0, getWidth(), getHeight());
                        }
                    }
                };
                // Tweaks for Look and Feel
                if (Util.isLAFAqua()) {
                    if (System.getProperty("java.version").compareTo("1.5.0_16")>=0) {
                        button.setMargin(new Insets(0,0,0,0));
                    } else {
                        button.setMargin(new Insets(4,5,5,5));
                    }
                } else if (Util.isLAFGTK()) {
                    button.setMargin(new Insets(0,0,0,0));
                } else if (Util.isLAFNimbus()) {
                    button.setMargin(new Insets(1,1,1,1));
                } else {
                    button.setMargin(new Insets(2,2,2,2));
                }
                String tt = UIManager.getString(tooltip);
                if (tt==null) {
                    if (viewer.getPropertyManager().getProperty("debug.L10N")!=null) {
                        System.out.println("No localization for \""+tooltip+"\"");
                    }
                    tt = tooltip;
                }
                button.setToolTipText(tt);
                button.setBorderPainted(false);
                button.setEnabled(!documentrequired);
                button.addActionListener(listener);
                component = button;
                compname = "Button"+compname;
            }

            boolean enabled = isButtonEnabledByDefault();
            if (propertymanager.getProperty(property+".button")!=null) {
                enabled = !propertymanager.getProperty(property+".button").equals("false");
            }
            if (enabled) {
                JComponent c = toolbar;
                if (toolbar instanceof JInternalFrame) {
                    c = (JComponent)((JInternalFrame)c).getContentPane();
                }
                if (ltrsubpanel) {
                    if (c.getComponentCount() == 0 || !(c.getComponent(0) instanceof JToolBar)) {
                        final JToolBar ltr = new JToolBar();
                        ltr.setLayout(new GridBagLayout() {
                            { defaultConstraints.fill = GridBagConstraints.BOTH; defaultConstraints.weighty = 1; }
                        });
                        ltr.setOpaque(false);
                        ltr.setBorder(BorderFactory.createEmptyBorder());
                        ltr.setFloatable(false);
                        ltr.addPropertyChangeListener(new PropertyChangeListener() {
                            public void propertyChange(PropertyChangeEvent event) {
                                if (event.getPropertyName().equals("componentOrientation")) {
                                    ComponentOrientation orient = (ComponentOrientation)event.getNewValue();
                                    if (orient != null && !orient.isLeftToRight()) {
                                        ltr.applyComponentOrientation(ComponentOrientation.LEFT_TO_RIGHT);
                                    }
                                }
                            }
                        });
                        c.add(ltr);
                        c = ltr;
                    } else {
                        c = (JComponent)c.getComponent(0);
                    }
                }
                c.add(component);
                if (toolbar instanceof JInternalFrame) {
                    ((JInternalFrame)toolbar).pack();
                }
            }

            viewer.putNamedComponent(compname, component);
            if (documentrequired) {
                viewer.addDocumentDependentComponent(component);
            }
        }
        boolean enabled = isMenuEnabledByDefault();
        if (propertymanager.getProperty(property+".menu")!=null) {
            enabled = !propertymanager.getProperty(property+".menu").equals("false");
        }
        if (menu!=null && enabled) {
            JMenuItem item = viewer.setMenu(menu, mnemonic, documentrequired, listener);
            if (component==null) component = item;
            viewer.putNamedComponent("Menu"+getName(), item);
        }
    }

    /**
     * Get the Viewer this Feature has been added to.
     */
    public final PDFViewer getViewer() {
        return viewer;
    }

    /**
     * Set whether this feature requires a PDF to be loaded. Most
     * features except for the "Open" widget do, so the default is "true"
     */
    protected final void setDocumentRequired(boolean required) {
        this.documentrequired = required;
    }

    /**
     * Return whether this widget should be inactive if no Document is
     * selected.
     * @see #setDocumentRequired
     * @since 2.11.7
     */
    public boolean isDocumentRequired() {
        return this.documentrequired;
    }

    /**
     * Set a custom component to be displayed in the ToolBar for this feature.
     * @param toolbar the name of the toolbar to put the component in
     * @param component the component
     */
    protected final void setComponent(String toolbar, JComponent component) {
        this.toolbarname = toolbar;
        this.component = component;
    }

    /**
     * Return the component representing this Widget.
     * @since 2.8.5
     */
    public JComponent getComponent() {
        return component;
    }

    /**
     * Return true if the button component for this widget is enabled by default.
     * The default is "true"
     * @since 2.10.3
     */
    public boolean isButtonEnabledByDefault() {
        return true;
    }

    /**
     * Return true if the menu component for this widget is enabled by default.
     * The default is "true"
     * @since 2.10.3
     */
    public boolean isMenuEnabledByDefault() {
        return true;
    }

    /**
     * Set this feature to use a regular button in the toolbar. The button will be
     * created using the specified icon and with the specified tooltip.
     * @param toolbar the name of the toolbar to put the component in
     * @param icon the URL of the icon to use
     * @param tooltip the tooltip to display for this button
     */
    protected final void setButton(String toolbar, String icon, String tooltip) {
        this.toolbarname = toolbar;
        this.icon = icon;
        this.tooltip = tooltip;
    }

    /**
     * Set whether the toolbar this feature is stored in is enabled by default
     */
    protected final void setToolBarEnabled(boolean enabled) {
        this.toolbarenabled = enabled;
    }

    /**
     * Set whether the toolbar this feature is stored in can be enabled or disabled
     * @see ToolbarDisabling
     */
    protected final void setToolBarEnabledAlways(boolean always) {
        this.toolbarenabledalways = always;
    }

    /**
     * Set whether the toolbar this feature is stored in can be floated
     */
    protected final void setToolBarFloatable(boolean floatable) {
        this.toolbarfloatable = floatable;
    }

    /**
     * Set whether this toolbar is always floating or not. Toolbars with this set
     * are implemented as JInternalFrame objects, and are never attached to the
     * regular tool bar
     * @since 2.8.3
     */
    protected final void setToolBarFloating(boolean floating) {
        this.floating = floating;
    }

    /**
     * Set a menu item for this feature. Activating the menu item is the
     * same as pressing the button.
     * @param menu the menu hierarchy to use, separated with tabs - eg "File\tOpen"
     */
    protected final void setMenu(String menu) {
        setMenu(menu, (char)0);
    }

    /**
     * Set a menu item for this feature, with an optional keyboard shortcut.
     * @param menu the menu hierarchy to use, separated with tabs - eg "File\tOpen"
     * @param mnemonic the keyboard shortcut to activate the menu - a lowercase or uppercase
     * character to activate the menu.
     * @since 2.10.2
     */
    protected final void setMenu(String menu, char mnemonic) {
        this.menu = menu;
        this.mnemonic = mnemonic;
    }

    /**
     * The method that's run when this feature is activated
     */
    public void action(ViewerEvent event) { }
}
